@@ -3,23 +3,9 @@
  *
  * RSyntaxDocument.java - A document capable of syntax highlighting, used by
  * RSyntaxTextArea.
- * Copyright (C) 2004 Robert Futrell
- * robert_futrell at users.sourceforge.net
- * http://fifesoft.com/rsyntaxtextarea
- *
- * This library is free software; you can redistribute it and/or
- * modify it under the terms of the GNU Lesser General Public
- * License as published by the Free Software Foundation; either
- * version 2.1 of the License, or (at your option) any later version.
- *
- * This library is distributed in the hope that it will be useful,
- * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
- * Lesser General Public License for more details.
- *
- * You should have received a copy of the GNU Lesser General Public
- * License along with this library; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA.
+ * 
+ * This library is distributed under a modified BSD license.  See the included
+ * RSyntaxTextArea.License.txt file for details.
  */
 package org.fife.ui.rsyntaxtextarea;
 
@@ -31,6 +17,7 @@ import javax.swing.event.*;
 import javax.swing.text.*;
 
 import org.fife.ui.rsyntaxtextarea.modes.AbstractMarkupTokenMaker;
+import org.fife.ui.rtextarea.RDocument;
 import org.fife.util.DynamicIntArray;
 
 
@@ -52,12 +39,12 @@ import org.fife.util.DynamicIntArray;
  * type <code>CHANGE</code> use their offset and length values to represent the
  * first and last lines, respectively, that have had their syntax coloring
  * change.  This is really a hack to increase the speed of the painting code
- * and should really be corrected, but oh well.
+ * and should really be corrected, but oh well. 
  *
  * @author Robert Futrell
- * @version 0.1
+ * @version 1.0
  */
-public class RSyntaxDocument extends PlainDocument implements SyntaxConstants {
+public class RSyntaxDocument extends RDocument implements SyntaxConstants {
 
 	/**
 	 * Creates a {@link TokenMaker} appropriate for a given programming
@@ -68,19 +55,35 @@ public class RSyntaxDocument extends PlainDocument implements SyntaxConstants {
 	/**
 	 * Splits text into tokens for the current programming language.
 	 */
-	private TokenMaker tokenMaker;
+	private transient TokenMaker tokenMaker;
+
+	/**
+	 * The current syntax style.  Only cached to keep this class serializable.
+	 */
+	private String syntaxStyle;
 
 	/**
 	 * Array of values representing the "last token type" on each line.  This
-	 * is used in cases such as multiline comments:  if the previous line
-	 * ended with an (unclosed) multiline comment, we can use this knowledge
-	 * and start the current line's syntax highlighting in multiline comment
+	 * is used in cases such as multi-line comments:  if the previous line
+	 * ended with an (unclosed) multi-line comment, we can use this knowledge
+	 * and start the current line's syntax highlighting in multi-line comment
 	 * state.
 	 */
-	protected DynamicIntArray lastTokensOnLines;
+	protected transient DynamicIntArray lastTokensOnLines;
+
+	private transient int lastLine = -1;
+	private transient Token cachedTokenList;
+	private transient int useCacheCount = 0;
+	private transient int tokenRetrievalCount = 0;
 
 	private transient Segment s;
 
+	/**
+	 * If this is set to <code>true</code>, debug information about how much
+	 * token caching is helping is printed to stdout.
+	 */
+	private static final boolean DEBUG_TOKEN_CACHING = false;
+
 
 	/**
 	 * Constructs a plain text document.  A default root element is created,
@@ -102,8 +105,7 @@ public class RSyntaxDocument extends PlainDocument implements SyntaxConstants {
 	 * @param syntaxStyle The syntax highlighting scheme to use.
 	 */
 	public RSyntaxDocument(TokenMakerFactory tmf, String syntaxStyle) {
-		super(new RGapContent());
-		putProperty(tabSizeAttribute, new Integer(5));
+		putProperty(tabSizeAttribute, Integer.valueOf(5));
 		lastTokensOnLines = new DynamicIntArray(400);
 		lastTokensOnLines.add(Token.NULL); // Initial (empty) line.
 		s = new Segment();
@@ -112,18 +114,6 @@ public class RSyntaxDocument extends PlainDocument implements SyntaxConstants {
 	}
 
 
-	/**
-	 * Returns the character in the document at the specified offset.
-	 *
-	 * @param offset The offset of the character.
-	 * @return The character.
-	 * @throws BadLocationException If the offset is invalid.
-	 */
-	public char charAt(int offset) throws BadLocationException {
-		return ((RGapContent)getContent()).charAt(offset);
-	}
-
-
 	/**
 	 * Alerts all listeners to this document of an insertion.  This is
 	 * overridden so we can update our syntax highlighting stuff.<p>
@@ -133,8 +123,11 @@ public class RSyntaxDocument extends PlainDocument implements SyntaxConstants {
 	 *
 	 * @param e The change.
 	 */
+	@Override
 	protected void fireInsertUpdate(DocumentEvent e) {
 
+		cachedTokenList = null;
+
 		/*
 		 * Now that the text is actually inserted into the content and
 		 * element structure, we can update our token elements and "last
@@ -205,8 +198,10 @@ public class RSyntaxDocument extends PlainDocument implements SyntaxConstants {
 	 * @param chng The change that occurred.
 	 * @see #removeUpdate
 	 */
+	@Override
 	protected void fireRemoveUpdate(DocumentEvent chng) {
 
+		cachedTokenList =  null;
 		Element lineMap = getDefaultRootElement();
 		int numLines = lineMap.getElementCount();
 
@@ -259,6 +254,19 @@ public class RSyntaxDocument extends PlainDocument implements SyntaxConstants {
 	}
 
 
+	/**
+	 * Returns the closest {@link TokenTypes "standard" token type} for a given
+	 * "internal" token type (e.g. one whose value is <code>&lt; 0</code>).
+	 *
+	 * @param type The token type.
+	 * @return The closest "standard" token type.  If a mapping is not defined
+	 *         for this language, then <code>type</code> is returned.
+	 */
+	public int getClosestStandardTokenTypeForInternalType(int type) {
+		return tokenMaker.getClosestStandardTokenTypeForInternalType(type);
+	}
+
+
 	/**
 	 * Returns whether closing markup tags should be automatically completed.
 	 * This method only returns <code>true</code> if
@@ -336,6 +344,16 @@ public class RSyntaxDocument extends PlainDocument implements SyntaxConstants {
 	}
 
 
+	/**
+	 * Returns the occurrence marker for the current language.
+	 *
+	 * @return The occurrence marker.
+	 */
+	OccurrenceMarker getOccurrenceMarker() {
+		return tokenMaker.getOccurrenceMarker();
+	}
+
+
 	/**
 	 * This method returns whether auto indentation should be done if Enter
 	 * is pressed at the end of the specified line.
@@ -360,6 +378,18 @@ public class RSyntaxDocument extends PlainDocument implements SyntaxConstants {
 	 * @return A token list representing the specified line.
 	 */
 	public final Token getTokenListForLine(int line) {
+
+		tokenRetrievalCount++;
+		if (line==lastLine && cachedTokenList!=null) {
+			if (DEBUG_TOKEN_CACHING) {
+				useCacheCount++;
+				System.err.println("--- Using cached line; ratio now: " +
+						useCacheCount + "/" + tokenRetrievalCount);
+			}
+			return cachedTokenList;
+		}
+		lastLine = line;
+		
 		Element map = getDefaultRootElement();
 		Element elem = map.getElement(line);
 		int startOffset = elem.getStartOffset();
@@ -374,7 +404,11 @@ public class RSyntaxDocument extends PlainDocument implements SyntaxConstants {
 		}
 		int initialTokenType = line==0 ? Token.NULL :
 								getLastTokenTypeOnLine(line-1);
-		return tokenMaker.getTokenList(s, initialTokenType, startOffset);
+
+		//return tokenMaker.getTokenList(s, initialTokenType, startOffset);
+		cachedTokenList = tokenMaker.getTokenList(s, initialTokenType, startOffset);
+		return cachedTokenList;
+
 	}
 
 
@@ -391,14 +425,26 @@ public class RSyntaxDocument extends PlainDocument implements SyntaxConstants {
 	/**
 	 * Deserializes a document.
 	 *
-	 * @param s The stream to read from.
+	 * @param in The stream to read from.
 	 * @throws ClassNotFoundException
 	 * @throws IOException
 	 */
-	private void readObject(ObjectInputStream s)
+	private void readObject(ObjectInputStream in)
 						throws ClassNotFoundException, IOException {
-		s.defaultReadObject();
+
+		in.defaultReadObject();
+
+		// Install default TokenMakerFactory.  To support custom TokenMakers,
+		// both JVM's should install default TokenMakerFactories that support
+		// the language they want to use beforehand.
+		setTokenMakerFactory(null);
+
+		// Handle other transient stuff
 		this.s = new Segment();
+		int lineCount = getDefaultRootElement().getElementCount();
+		lastTokensOnLines = new DynamicIntArray(lineCount);
+		setSyntaxStyle(syntaxStyle); // Actually install (transient) TokenMaker
+
 	}
 
 
@@ -441,10 +487,12 @@ public class RSyntaxDocument extends PlainDocument implements SyntaxConstants {
 	 *        {@link SyntaxConstants#SYNTAX_STYLE_JAVA}.  If this style is not
 	 *        known or supported by this document, then
 	 *        {@link SyntaxConstants#SYNTAX_STYLE_NONE} is used.
+	 * @see #setSyntaxStyle(TokenMaker)
 	 */
 	public void setSyntaxStyle(String styleKey) {
 		tokenMaker = tokenMakerFactory.getTokenMaker(styleKey);
 		updateSyntaxHighlightingInformation();
+		this.syntaxStyle = styleKey;
 	}
 
 
@@ -455,6 +503,7 @@ public class RSyntaxDocument extends PlainDocument implements SyntaxConstants {
 	 * <code>RSyntaxTextArea</code>.
 	 *
 	 * @param tokenMaker The new token maker to use.
+	 * @see #setSyntaxStyle(String)
 	 */
 	public void setSyntaxStyle(TokenMaker tokenMaker) {
 		this.tokenMaker = tokenMaker;
@@ -474,18 +523,6 @@ public class RSyntaxDocument extends PlainDocument implements SyntaxConstants {
 	}
 
 
-	/**
-	 * Sets whether whitespace is visible.  This property is actually setting
-	 * whether the tokens generated from this document "paint" something when
-	 * they represent whitespace.
-	 *
-	 * @param visible Whether whitespace should be visible.
-	 */
-	public void setWhitespaceVisible(boolean visible, RSyntaxTextArea textArea) {
-		tokenMaker.setWhitespaceVisible(visible, textArea);
-	}
-
-
 	/**
 	 * Loops through the last-tokens-on-lines array from a specified point
 	 * onward, updating last-token values until they stop changing.  This
@@ -581,30 +618,4 @@ public class RSyntaxDocument extends PlainDocument implements SyntaxConstants {
 	}
 
 
-	/**
-	 * Document content that provides access to individual characters.
-	 *
-	 * @author Robert Futrell
-	 * @version 1.0
-	 */
-	private static class RGapContent extends GapContent {
-
-		public RGapContent() {
-		}
-
-		public char charAt(int offset) throws BadLocationException {
-			if (offset<0 || offset>=length()) {
-				throw new BadLocationException("Invalid offset", offset);
-			}
-			int g0 = getGapStart();
-			char[] array = (char[]) getArray();
-			if (offset<g0) { // below gap
-				return array[offset];
-			}
-			return array[getGapEnd() + offset - g0]; // above gap
-		}
-
-	}
-
-
 }
\ No newline at end of file
