@@ -3,23 +3,9 @@
  *
  * SyntaxScheme.java - The set of colors and tokens used by an RSyntaxTextArea
  * to color tokens.
- * Copyright (C) 2004 Robert Futrell
- * robert_futrell at users.sourceforge.net
- * http://fifesoft.com/rsyntaxtextarea
- *
- * This library is free software; you can redistribute it and/or
- * modify it under the terms of the GNU Lesser General Public
- * License as published by the Free Software Foundation; either
- * version 2.1 of the License, or (at your option) any later version.
- *
- * This library is distributed in the hope that it will be useful,
- * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
- * Lesser General Public License for more details.
- *
- * You should have received a copy of the GNU Lesser General Public
- * License along with this library; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA.
+ * 
+ * This library is distributed under a modified BSD license.  See the included
+ * RSyntaxTextArea.License.txt file for details.
  */
 package org.fife.ui.rsyntaxtextarea;
 
@@ -40,14 +26,18 @@ import org.xml.sax.helpers.XMLReaderFactory;
 
 /**
  * The set of colors and styles used by an <code>RSyntaxTextArea</code> to
- * color tokens.
+ * color tokens.<p>
+ * You can use this class to programmatically set the fonts and colors used in
+ * an RSyntaxTextArea, but for more powerful, externalized control, consider
+ * using {@link Theme}s instead.
  *
  * @author Robert Futrell
  * @version 1.0
+ * @see Theme
  */
-public class SyntaxScheme implements Cloneable {
+public class SyntaxScheme implements Cloneable, TokenTypes {
 
-	public Style[] styles;
+	private Style[] styles;
 
 	private static final String VERSION			= "*ver1";
 
@@ -61,7 +51,7 @@ public class SyntaxScheme implements Cloneable {
 	 *        will be initially <code>null</code>.
 	 */
 	public SyntaxScheme(boolean useDefaults) {
-		styles = new Style[Token.NUM_TOKEN_TYPES];
+		styles = new Style[DEFAULT_NUM_TOKEN_TYPES];
 		if (useDefaults) {
 			restoreDefaults(null);
 		}
@@ -76,8 +66,22 @@ public class SyntaxScheme implements Cloneable {
 	 *        font.
 	 */
 	public SyntaxScheme(Font baseFont) {
-		styles = new Style[Token.NUM_TOKEN_TYPES];
-		restoreDefaults(baseFont);
+		this(baseFont, true);
+	}
+
+
+	/**
+	 * Creates a default color scheme.
+	 *
+	 * @param baseFont The base font to use.  Keywords will be a bold version
+	 *        of this font, and comments will be an italicized version of this
+	 *        font.
+	 * @param fontStyles Whether bold and italic should be used in the scheme
+	 *        (vs. all tokens using a plain font).
+	 */
+	public SyntaxScheme(Font baseFont, boolean fontStyles) {
+		styles = new Style[DEFAULT_NUM_TOKEN_TYPES];
+		restoreDefaults(baseFont, fontStyles);
 	}
 
 
@@ -113,6 +117,7 @@ public class SyntaxScheme implements Cloneable {
 	 *
 	 * @return The copy.
 	 */
+	@Override
 	public Object clone() {
 		SyntaxScheme shcs = null;
 		try {
@@ -121,8 +126,8 @@ public class SyntaxScheme implements Cloneable {
 			cnse.printStackTrace();
 			return null;
 		}
-		shcs.styles = new Style[Token.NUM_TOKEN_TYPES];
-		for (int i=0; i<Token.NUM_TOKEN_TYPES; i++) {
+		shcs.styles = new Style[styles.length];
+		for (int i=0; i<styles.length; i++) {
 			Style s = styles[i];
 			if (s!=null) {
 				shcs.styles[i] = (Style)s.clone();
@@ -140,6 +145,7 @@ public class SyntaxScheme implements Cloneable {
 	 *         <code>otherScheme</code> are the same scheme;
 	 *         <code>false</code> otherwise.
 	 */
+	@Override
 	public boolean equals(Object otherScheme) {
 
 		// No need for null check; instanceof takes care of this for us,
@@ -179,12 +185,53 @@ public class SyntaxScheme implements Cloneable {
 	}
 
 
+	/**
+	 * Returns the specified style.
+	 *
+	 * @param index The index of the style.
+	 * @return The style.
+	 * @see #setStyle(int, Style)
+	 * @see #getStyleCount()
+	 */
+	public Style getStyle(int index) {
+		return styles[index];
+	}
+
+
+	/**
+	 * Returns the number of styles.
+	 *
+	 * @return The number of styles.
+	 * @see #getStyle(int)
+	 */
+	public int getStyleCount() {
+		return styles.length;
+	}
+
+
+	/**
+	 * Used by third party implementors e.g. SquirreL SQL. Most applications do
+	 * not need to call this method.
+	 * <p>
+	 * Note that the returned array is not a copy of the style data; editing the
+	 * array will modify the styles used by any <code>RSyntaxTextArea</code>
+	 * using this scheme.
+	 *
+	 * @return The style array.
+	 * @see #setStyles(Style[])
+	 */
+	public Style[] getStyles() {
+		return styles;
+	}
+
+
 	/**
 	 * This is implemented to be consistent with {@link #equals(Object)}.
 	 * This is a requirement to keep FindBugs happy.
 	 *
 	 * @return The hash code for this object.
 	 */
+	@Override
 	public int hashCode() {
 		// Keep me fast.  Iterating over *all* syntax schemes contained is
 		// probably much slower than a "bad" hash code here.
@@ -201,7 +248,10 @@ public class SyntaxScheme implements Cloneable {
 
 
 	/**
-	 * Loads a syntax scheme from an input stream.
+	 * Loads a syntax scheme from an input stream.<p>
+	 * 
+	 * Consider using the {@link Theme} class for saving and loading RSTA
+	 * styles rather than using this API.
 	 *
 	 * @param baseFont The font to use as the "base" for the syntax scheme.
 	 *        If this is <code>null</code>, a default monospaced font is used.
@@ -215,19 +265,47 @@ public class SyntaxScheme implements Cloneable {
 		if (baseFont==null) {
 			baseFont = RSyntaxTextArea.getDefaultFont();
 		}
-		return XmlParser.load(baseFont, in);
+		return SyntaxSchemeLoader.load(baseFont, in);
 	}
 
 
 	/**
 	 * Loads a syntax highlighting color scheme from a string created from
 	 * <code>toCommaSeparatedString</code>.  This method is useful for saving
-	 * and restoring color schemes.
+	 * and restoring color schemes.<p>
+	 * 
+	 * Consider using the {@link Theme} class for saving and loading RSTA
+	 * styles rather than using this API.
 	 *
 	 * @param string A string generated from {@link #toCommaSeparatedString()}.
 	 * @return A color scheme.
+	 * @see #toCommaSeparatedString()
 	 */
 	public static SyntaxScheme loadFromString(String string) {
+		return loadFromString(string, DEFAULT_NUM_TOKEN_TYPES);
+	}
+
+
+	/**
+	 * Loads a syntax highlighting color scheme from a string created from
+	 * <code>toCommaSeparatedString</code>.  This method is useful for saving
+	 * and restoring color schemes.<p>
+	 * 
+	 * Consider using the {@link Theme} class for saving and loading RSTA
+	 * styles rather than using this API.
+	 *
+	 * @param string A string generated from {@link #toCommaSeparatedString()}.
+	 * @param tokenTypeCount The number of token types saved in this string.
+	 *        This should be the number of token types saved by your custom
+	 *        SyntaxScheme subclass,
+	 *        or {@link TokenTypes#DEFAULT_NUM_TOKEN_TYPES} if you used the
+	 *        standard implementation (which most people will).
+	 * @return A color scheme.
+	 * @see #loadFromString(String)
+	 * @see #toCommaSeparatedString()
+	 */
+	public static SyntaxScheme loadFromString(String string,
+			int tokenTypeCount) {
 
 		SyntaxScheme scheme = new SyntaxScheme(true);
 
@@ -242,7 +320,6 @@ public class SyntaxScheme implements Cloneable {
 					return scheme; // Still set to defaults
 				}
 
-				int tokenTypeCount = Token.NUM_TOKEN_TYPES;
 				int tokenCount = tokenTypeCount*7 + 1; // Version string
 				if (tokens.length!=tokenCount) {
 					throw new Exception(
@@ -250,6 +327,10 @@ public class SyntaxScheme implements Cloneable {
 						tokenCount + ", found " + tokens.length);
 				}
 
+				// Use StyleContext to create fonts to get composite fonts for
+				// Asian glyphs.
+				StyleContext sc = StyleContext.getDefaultStyleContext();
+
 				// Loop through each token style.  Format:
 				// "index,(fg|-),(bg|-),(t|f),((font,style,size)|(-,,))"
 				for (int i=0; i<tokenTypeCount; i++) {
@@ -280,7 +361,7 @@ public class SyntaxScheme implements Cloneable {
 					Font font = null;
 					String family = tokens[pos+4];
 					if (!"-".equals(family)) {
-						font = new Font(family,
+						font = sc.getFont(family,
 							Integer.parseInt(tokens[pos+5]),	// style
 							Integer.parseInt(tokens[pos+6]));	// size
 					}
@@ -319,56 +400,92 @@ public class SyntaxScheme implements Cloneable {
 	 *        used.
 	 */
 	public void restoreDefaults(Font baseFont) {
+		restoreDefaults(baseFont, true);
+	}
+
+
+	/**
+	 * Restores all colors and fonts to their default values.
+	 *
+	 * @param baseFont The base font to use when creating this scheme.  If
+	 *        this is <code>null</code>, then a default monospaced font is
+	 *        used.
+	 * @param fontStyles Whether bold and italic should be used in the scheme
+	 *        (vs. all tokens using a plain font).
+	 */
+	public void restoreDefaults(Font baseFont, boolean fontStyles) {
 
 		// Colors used by tokens.
 		Color comment			= new Color(0,128,0);
 		Color docComment		= new Color(164,0,0);
+		Color markupComment		= new Color(0, 96, 0);
 		Color keyword			= Color.BLUE;
+		Color dataType			= new Color(0,128,128);
 		Color function			= new Color(173,128,0);
+		Color preprocessor		= new Color(128,128,128);
+		Color operator			= new Color(128, 64, 64);
+		Color regex				= new Color(0,128,164);
+		Color variable			= new Color(255,153,0);
 		Color literalNumber		= new Color(100,0,200);
 		Color literalString		= new Color(220,0,156);
 		Color error			= new Color(148,148,0);
 
-		// Special fonts.
+		// (Possible) special font styles for keywords and comments.
 		if (baseFont==null) {
 			baseFont = RSyntaxTextArea.getDefaultFont();
 		}
-		// WORKAROUND for Sun JRE bug 6282887 (Asian font bug in 1.4/1.5)
-		StyleContext sc = StyleContext.getDefaultStyleContext();
-		Font boldFont = sc.getFont(baseFont.getFamily(), Font.BOLD,
-				baseFont.getSize());
-		Font italicFont = sc.getFont(baseFont.getFamily(), Font.ITALIC,
-				baseFont.getSize());
-		Font commentFont = italicFont;//baseFont.deriveFont(Font.ITALIC);
-		Font keywordFont = boldFont;//baseFont.deriveFont(Font.BOLD);
+		Font commentFont = baseFont;
+		Font keywordFont = baseFont;
+		if (fontStyles) {
+			// WORKAROUND for Sun JRE bug 6282887 (Asian font bug in 1.4/1.5)
+			// That bug seems to be hidden now, see 6289072 instead.
+			StyleContext sc = StyleContext.getDefaultStyleContext();
+			Font boldFont = sc.getFont(baseFont.getFamily(), Font.BOLD,
+					baseFont.getSize());
+			Font italicFont = sc.getFont(baseFont.getFamily(), Font.ITALIC,
+					baseFont.getSize());
+			commentFont = italicFont;//baseFont.deriveFont(Font.ITALIC);
+			keywordFont = boldFont;//baseFont.deriveFont(Font.BOLD);
+		}
 
-		styles[Token.COMMENT_EOL]				= new Style(comment, null, commentFont);
-		styles[Token.COMMENT_MULTILINE]			= new Style(comment, null, commentFont);
-		styles[Token.COMMENT_DOCUMENTATION]		= new Style(docComment, null, commentFont);
-		styles[Token.RESERVED_WORD]				= new Style(keyword, null, keywordFont);
-		styles[Token.FUNCTION]					= new Style(function, null);
-		styles[Token.LITERAL_BOOLEAN]			= new Style(literalNumber, null);
-		styles[Token.LITERAL_NUMBER_DECIMAL_INT]	= new Style(literalNumber, null);
-		styles[Token.LITERAL_NUMBER_FLOAT]		= new Style(literalNumber, null);
-		styles[Token.LITERAL_NUMBER_HEXADECIMAL]	= new Style(literalNumber, null);
-		styles[Token.LITERAL_STRING_DOUBLE_QUOTE]	= new Style(literalString, null);
-		styles[Token.LITERAL_CHAR]				= new Style(literalString, null);
-		styles[Token.LITERAL_BACKQUOTE]			= new Style(literalString, null);
-		styles[Token.DATA_TYPE]				= new Style(new Color(0,128,128), null);
-		styles[Token.VARIABLE]					= new Style(new Color(255,153,0), null);
-		styles[Token.IDENTIFIER]				= new Style(null, null);
-		styles[Token.WHITESPACE]				= new Style(Color.gray, null);
-		styles[Token.SEPARATOR]				= new Style(Color.RED, null);
-		styles[Token.OPERATOR]					= new Style(new Color(128,64,64), null);
-		styles[Token.PREPROCESSOR]				= new Style(new Color(128,128,128), null);
-		styles[Token.MARKUP_TAG_DELIMITER]		= new Style(Color.RED, null);
-		styles[Token.MARKUP_TAG_NAME]			= new Style(Color.BLUE, null);
-		styles[Token.MARKUP_TAG_ATTRIBUTE]		= new Style(new Color(63,127,127), null);
-//		styles[Token.ERROR]					= null;
-		styles[Token.ERROR_IDENTIFIER]			= new Style(error, null);
-		styles[Token.ERROR_NUMBER_FORMAT]		= new Style(error, null);
-		styles[Token.ERROR_STRING_DOUBLE]		= new Style(error, null);
-		styles[Token.ERROR_CHAR]				= new Style(error, null);
+		styles[COMMENT_EOL]				= new Style(comment, null, commentFont);
+		styles[COMMENT_MULTILINE]			= new Style(comment, null, commentFont);
+		styles[COMMENT_DOCUMENTATION]		= new Style(docComment, null, commentFont);
+		styles[COMMENT_KEYWORD]			= new Style(new Color(255,152,0), null, commentFont);
+		styles[COMMENT_MARKUP]			= new Style(Color.gray, null, commentFont);
+		styles[RESERVED_WORD]				= new Style(keyword, null, keywordFont);
+		styles[RESERVED_WORD_2]			= new Style(keyword, null, keywordFont);
+		styles[FUNCTION]					= new Style(function);
+		styles[LITERAL_BOOLEAN]			= new Style(literalNumber);
+		styles[LITERAL_NUMBER_DECIMAL_INT]	= new Style(literalNumber);
+		styles[LITERAL_NUMBER_FLOAT]		= new Style(literalNumber);
+		styles[LITERAL_NUMBER_HEXADECIMAL]	= new Style(literalNumber);
+		styles[LITERAL_STRING_DOUBLE_QUOTE]	= new Style(literalString);
+		styles[LITERAL_CHAR]				= new Style(literalString);
+		styles[LITERAL_BACKQUOTE]			= new Style(literalString);
+		styles[DATA_TYPE]				= new Style(dataType, null, keywordFont);
+		styles[VARIABLE]					= new Style(variable);
+		styles[REGEX]						= new Style(regex);
+		styles[ANNOTATION]				= new Style(Color.gray);
+		styles[IDENTIFIER]				= new Style(null);
+		styles[WHITESPACE]				= new Style(Color.gray);
+		styles[SEPARATOR]				= new Style(Color.RED);
+		styles[OPERATOR]					= new Style(operator);
+		styles[PREPROCESSOR]				= new Style(preprocessor);
+		styles[MARKUP_TAG_DELIMITER]		= new Style(Color.RED);
+		styles[MARKUP_TAG_NAME]			= new Style(Color.BLUE);
+		styles[MARKUP_TAG_ATTRIBUTE]		= new Style(new Color(63,127,127));
+		styles[MARKUP_TAG_ATTRIBUTE_VALUE]= new Style(literalString);
+		styles[MARKUP_COMMENT]              = new Style(markupComment, null, commentFont);
+		styles[MARKUP_DTD]              = new Style(function);
+		styles[MARKUP_PROCESSING_INSTRUCTION] = new Style(preprocessor);
+		styles[MARKUP_CDATA]				= new Style(new Color(0xcc6600));
+		styles[MARKUP_CDATA_DELIMITER]		= new Style(new Color(0x008080));
+		styles[MARKUP_ENTITY_REFERENCE]		= new Style(dataType);
+		styles[ERROR_IDENTIFIER]			= new Style(error);
+		styles[ERROR_NUMBER_FORMAT]		= new Style(error);
+		styles[ERROR_STRING_DOUBLE]		= new Style(error);
+		styles[ERROR_CHAR]				= new Style(error);
 
 	}
 
@@ -378,12 +495,29 @@ public class SyntaxScheme implements Cloneable {
 	 *
 	 * @param type The token type.
 	 * @param style The style for the token type.
+	 * @see #getStyle(int)
 	 */
 	public void setStyle(int type, Style style) {
 		styles[type] = style;
 	}
 
 
+	/**
+	 * Used by third party implementors e.g. SquirreL SQL. Most applications do
+	 * not need to call this method; individual styles can be set via
+	 * {@link #setStyle(int, Style)}.
+	 *
+	 * @param styles The new array of styles to use.  Note that this should
+	 *        have length of at least
+	 *        {@link TokenTypes#DEFAULT_NUM_TOKEN_TYPES}.
+	 * @see #setStyle(int, Style)
+	 * @see #getStyles()
+	 */
+	public void setStyles(Style[] styles) {
+		this.styles = styles;
+	}
+
+
 	/**
 	 * Returns the color represented by a string.  If the first char in the
 	 * string is '<code>$</code>', it is assumed to be in hex, otherwise it is
@@ -432,13 +566,14 @@ public class SyntaxScheme implements Cloneable {
 	 * </ul>
 	 *
 	 * @return A string representing the rgb values of the colors.
+	 * @see #loadFromString(String)
 	 */
 	public String toCommaSeparatedString() {
 
-		StringBuffer sb = new StringBuffer(VERSION);
+		StringBuilder sb = new StringBuilder(VERSION);
 		sb.append(',');
 
-		for (int i=0; i<Token.NUM_TOKEN_TYPES; i++) {
+		for (int i=0; i<styles.length; i++) {
 
 			sb.append(i).append(',');
 
@@ -471,50 +606,28 @@ public class SyntaxScheme implements Cloneable {
 
 	}
 
-
 	/**
 	 * Loads a <code>SyntaxScheme</code> from an XML file.
 	 */
-	private static class XmlParser extends DefaultHandler {
+	private static class SyntaxSchemeLoader extends DefaultHandler {
 
 		private Font baseFont;
 		private SyntaxScheme scheme;
 
-		public XmlParser(Font baseFont) {
+		public SyntaxSchemeLoader(Font baseFont) {
 			scheme = new SyntaxScheme(baseFont);
 		}
 
-		/**
-		 * Creates the XML reader to use.  Note that in 1.4 JRE's, the reader
-		 * class wasn't defined by default, but in 1.5+ it is.
-		 *
-		 * @return The XML reader to use.
-		 */
-		private static XMLReader createReader() throws IOException {
-			XMLReader reader = null;
-			try {
-				reader = XMLReaderFactory.createXMLReader();
-			} catch (SAXException e) {
-				// Happens in JRE 1.4.x; 1.5+ define the reader class properly
-				try {
-					reader = XMLReaderFactory.createXMLReader(
-							"org.apache.crimson.parser.XMLReaderImpl");
-				} catch (SAXException se) {
-					throw new IOException(se.toString());
-				}
-			}
-			return reader;
-		}
-
-		public static SyntaxScheme load(Font baseFont,
-										InputStream in) throws IOException {
-			XMLReader reader = createReader();
-			XmlParser parser = new XmlParser(baseFont);
-			parser.baseFont = baseFont;
-			reader.setContentHandler(parser);
-			InputSource is = new InputSource(in);
-			is.setEncoding("UTF-8");
+		public static SyntaxScheme load(Font baseFont, InputStream in)
+				throws IOException {
+			SyntaxSchemeLoader parser = null;
 			try {
+				XMLReader reader = XMLReaderFactory.createXMLReader();
+				parser = new SyntaxSchemeLoader(baseFont);
+				parser.baseFont = baseFont;
+				reader.setContentHandler(parser);
+				InputSource is = new InputSource(in);
+				is.setEncoding("UTF-8");
 				reader.parse(is);
 			} catch (SAXException se) {
 				throw new IOException(se.toString());
@@ -522,6 +635,7 @@ public class SyntaxScheme implements Cloneable {
 			return parser.scheme;
 		}
 
+		@Override
 		public void startElement(String uri, String localName, String qName,
 								Attributes attrs) {
 
