@@ -3,23 +3,9 @@
  *
  * Gutter.java - Manages line numbers, icons, etc. on the left-hand side of
  * an RTextArea.
- * Copyright (C) 2009 Robert Futrell
- * robert_futrell at users.sourceforge.net
- * http://fifesoft.com/rsyntaxtextarea
- *
- * This library is free software; you can redistribute it and/or
- * modify it under the terms of the GNU Lesser General Public
- * License as published by the Free Software Foundation; either
- * version 2.1 of the License, or (at your option) any later version.
- *
- * This library is distributed in the hope that it will be useful,
- * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
- * Lesser General Public License for more details.
- *
- * You should have received a copy of the GNU Lesser General Public
- * License along with this library; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA.
+ * 
+ * This library is distributed under a modified BSD license.  See the included
+ * RSyntaxTextArea.License.txt file for details.
  */
 package org.fife.ui.rtextarea;
 
@@ -30,11 +16,14 @@ import java.awt.ComponentOrientation;
 import java.awt.Font;
 import java.awt.Graphics;
 import java.awt.Point;
+import java.awt.Rectangle;
 import java.awt.event.ComponentAdapter;
 import java.beans.PropertyChangeEvent;
 import java.beans.PropertyChangeListener;
 import javax.swing.Icon;
 import javax.swing.JComponent;
+import javax.swing.JPanel;
+import javax.swing.border.Border;
 import javax.swing.border.EmptyBorder;
 import javax.swing.event.DocumentEvent;
 import javax.swing.event.DocumentListener;
@@ -43,26 +32,46 @@ import javax.swing.text.BadLocationException;
 import org.fife.ui.rsyntaxtextarea.ActiveLineRangeEvent;
 import org.fife.ui.rsyntaxtextarea.ActiveLineRangeListener;
 import org.fife.ui.rsyntaxtextarea.RSyntaxTextArea;
+import org.fife.ui.rsyntaxtextarea.folding.FoldManager;
 
 
 /**
  * The gutter is the component on the left-hand side of the text area that
- * displays optional information such as line numbers and icons (for bookmarks,
- * debugging breakpoints, error markers, etc.).<p>
+ * displays optional information such as line numbers, fold regions, and icons
+ * (for bookmarks, debugging breakpoints, error markers, etc.).<p>
  *
- * To add icons to the gutter, you must first call
- * {@link RTextScrollPane#setIconRowHeaderEnabled(boolean)} on the parent
- * scroll pane, to make the icon area visible.  Then, you can add icons that
- * track either lines in the document, or offsets, via
+ * Icons can be added on a per-line basis to visually mark syntax errors, lines
+ * with breakpoints set on them, etc.  To add icons to the gutter, you must
+ * first call {@link RTextScrollPane#setIconRowHeaderEnabled(boolean)} on the
+ * parent scroll pane, to make the icon area visible.  Then, you can add icons
+ * that track either lines in the document, or offsets, via
  * {@link #addLineTrackingIcon(int, Icon)} and
  * {@link #addOffsetTrackingIcon(int, Icon)}, respectively.  To remove an
- * icon you've added, use {@link #removeTrackingIcon(GutterIconInfo)}.
+ * icon you've added, use {@link #removeTrackingIcon(GutterIconInfo)}.<p>
+ *
+ * In addition to support for arbitrary per-line icons, this component also
+ * has built-in support for displaying icons representing "bookmarks;" that is,
+ * lines a user can cycle through via F2 and Shift+F2.  Bookmarked lines are
+ * toggled via Ctrl+F2.  In order to enable bookmarking, you must first assign
+ * an icon to represent a bookmarked line, then actually enable the feature:
+ * 
+ * <pre>
+ * Gutter gutter = scrollPane.getGutter();
+ * gutter.setBookmarkIcon(new ImageIcon("bookmark.png"));
+ * gutter.setBookmarkingEnabled(true);
+ * </pre>
  *
  * @author Robert Futrell
  * @version 1.0
  * @see GutterIconInfo
  */
-public class Gutter extends JComponent {
+public class Gutter extends JPanel {
+
+	/**
+	 * The color used to highlight active line ranges if none is specified.
+	 */
+	public static final Color DEFAULT_ACTIVE_LINE_RANGE_COLOR =
+												new Color(51, 153, 255);
 
 	/**
 	 * The text area.
@@ -74,11 +83,31 @@ public class Gutter extends JComponent {
 	 */
 	private LineNumberList lineNumberList;
 
+	/**
+	 * The color used to render line numbers.
+	 */
+	private Color lineNumberColor;
+
+	/**
+	 * The starting index for line numbers in the gutter.
+	 */
+	private int lineNumberingStartIndex;
+
+	/**
+	 * The font used to render line numbers.
+	 */
+	private Font lineNumberFont;
+
 	/**
 	 * Renders bookmark icons, breakpoints, error icons, etc.
 	 */
 	private IconRowHeader iconArea;
 
+	/**
+	 * Shows lines that are code-foldable.
+	 */
+	private FoldIndicator foldIndicator;
+
 	/**
 	 * Listens for events in our text area.
 	 */
@@ -93,12 +122,20 @@ public class Gutter extends JComponent {
 	public Gutter(RTextArea textArea) {
 
 		listener = new TextAreaListener();
+		lineNumberColor = Color.gray;
+		lineNumberFont = RTextArea.getDefaultFont();
+		lineNumberingStartIndex = 1;
+
 		setTextArea(textArea);
 		setLayout(new BorderLayout());
 		if (this.textArea!=null) {
 			// Enable line numbers our first time through if they give us
 			// a text area.
 			setLineNumbersEnabled(true);
+			if (this.textArea instanceof RSyntaxTextArea) {
+				RSyntaxTextArea rsta = (RSyntaxTextArea)this.textArea;
+				setFoldIndicatorEnabled(rsta.isCodeFoldingEnabled());
+			}
 		}
 
 		setBorder(new GutterBorder(0, 0, 0, 1)); // Assume ltr
@@ -124,13 +161,37 @@ public class Gutter extends JComponent {
 	 *         icon.
 	 * @throws BadLocationException If <code>offs</code> is an invalid offset
 	 *         into the text area.
+	 * @see #addLineTrackingIcon(int, Icon, String)
 	 * @see #addOffsetTrackingIcon(int, Icon)
 	 * @see #removeTrackingIcon(GutterIconInfo)
 	 */
 	public GutterIconInfo addLineTrackingIcon(int line, Icon icon)
 											throws BadLocationException {
+		return addLineTrackingIcon(line, icon, null);
+	}
+
+
+	/**
+	 * Adds an icon that tracks an offset in the document, and is displayed
+	 * adjacent to the line numbers.  This is useful for marking things such
+	 * as source code errors.
+	 *
+	 * @param line The line to track (zero-based).
+	 * @param icon The icon to display.  This should be small (say 16x16).
+	 * @param tip An optional tool tip for the icon.
+	 * @return A tag for this icon.  This can later be used in a call to
+	 *         {@link #removeTrackingIcon(GutterIconInfo)} to remove this
+	 *         icon.
+	 * @throws BadLocationException If <code>offs</code> is an invalid offset
+	 *         into the text area.
+	 * @see #addLineTrackingIcon(int, Icon)
+	 * @see #addOffsetTrackingIcon(int, Icon)
+	 * @see #removeTrackingIcon(GutterIconInfo)
+	 */
+	public GutterIconInfo addLineTrackingIcon(int line, Icon icon, String tip)
+											throws BadLocationException {
 		int offs = textArea.getLineStartOffset(line);
-		return addOffsetTrackingIcon(offs, icon);
+		return addOffsetTrackingIcon(offs, icon, tip);
 	}
 
 
@@ -144,12 +205,34 @@ public class Gutter extends JComponent {
 	 * @return A tag for this icon.
 	 * @throws BadLocationException If <code>offs</code> is an invalid offset
 	 *         into the text area.
+	 * @see #addOffsetTrackingIcon(int, Icon, String)
 	 * @see #addLineTrackingIcon(int, Icon)
 	 * @see #removeTrackingIcon(GutterIconInfo)
 	 */
 	public GutterIconInfo addOffsetTrackingIcon(int offs, Icon icon)
 												throws BadLocationException {
-		return iconArea.addOffsetTrackingIcon(offs, icon);
+		return addOffsetTrackingIcon(offs, icon, null);
+	}
+
+
+	/**
+	 * Adds an icon that tracks an offset in the document, and is displayed
+	 * adjacent to the line numbers.  This is useful for marking things such
+	 * as source code errors.
+	 *
+	 * @param offs The offset to track.
+	 * @param icon The icon to display.  This should be small (say 16x16).
+	 * @param tip An optional tool tip for the icon.
+	 * @return A tag for this icon.
+	 * @throws BadLocationException If <code>offs</code> is an invalid offset
+	 *         into the text area.
+	 * @see #addOffsetTrackingIcon(int, Icon)
+	 * @see #addLineTrackingIcon(int, Icon)
+	 * @see #removeTrackingIcon(GutterIconInfo)
+	 */
+	public GutterIconInfo addOffsetTrackingIcon(int offs, Icon icon, String tip)
+												throws BadLocationException {
+		return iconArea.addOffsetTrackingIcon(offs, icon, tip);
 	}
 
 
@@ -163,6 +246,17 @@ public class Gutter extends JComponent {
 	}
 
 
+	/**
+	 * Returns the color used to paint the active line range, if any.
+	 *
+	 * @return The color.
+	 * @see #setActiveLineRangeColor(Color)
+	 */
+	public Color getActiveLineRangeColor() {
+		return iconArea.getActiveLineRangeColor();
+	}
+
+
 	/**
 	 * Returns the icon to use for bookmarks.
 	 *
@@ -198,6 +292,28 @@ public class Gutter extends JComponent {
 	}
 
 
+	/**
+	 * Returns the background color used by the (default) fold icons.
+	 *
+	 * @return The background color.
+	 * @see #setFoldBackground(Color)
+	 */
+	public Color getFoldBackground() {
+		return foldIndicator.getFoldIconBackground();
+	}
+
+
+	/**
+	 * Returns the foreground color of the fold indicator.
+	 *
+	 * @return The foreground color of the fold indicator.
+	 * @see #setFoldIndicatorForeground(Color)
+	 */
+	public Color getFoldIndicatorForeground() {
+		return foldIndicator.getForeground();
+	}
+
+
 	/**
 	 * Returns the color to use to paint line numbers.
 	 *
@@ -205,7 +321,7 @@ public class Gutter extends JComponent {
 	 * @see #setLineNumberColor(Color)
 	 */
 	public Color getLineNumberColor() {
-		return lineNumberList.getForeground();
+		return lineNumberColor;
 	}
 
 
@@ -216,7 +332,7 @@ public class Gutter extends JComponent {
 	 * @see #setLineNumberFont(Font)
 	 */
 	public Font getLineNumberFont() {
-		return lineNumberList.getFont();
+		return lineNumberFont;
 	}
 
 
@@ -228,7 +344,7 @@ public class Gutter extends JComponent {
 	 * @see #setLineNumberingStartIndex(int)
 	 */
 	public int getLineNumberingStartIndex() {
-		return lineNumberList.getLineNumberingStartIndex();
+		return lineNumberingStartIndex;
 	}
 
 
@@ -247,6 +363,18 @@ public class Gutter extends JComponent {
 	}
 
 
+	/**
+	 * Returns whether tool tips are displayed showing the contents of
+	 * collapsed fold regions when the mouse hovers over a +/- icon.
+	 *
+	 * @return Whether these tool tips are displayed.
+	 * @see #setShowCollapsedRegionToolTips(boolean)
+	 */
+	public boolean getShowCollapsedRegionToolTips() {
+		return foldIndicator.getShowCollapsedRegionToolTips();
+	}
+
+
 	/**
 	 * Returns the tracking icons at the specified view position.
 	 *
@@ -255,13 +383,30 @@ public class Gutter extends JComponent {
 	 *         icons there, this will be an empty array.
 	 * @throws BadLocationException If <code>p</code> is invalid.
 	 */
-	public Object[] getTrackingIcons(Point p) throws BadLocationException {
+	public GutterIconInfo[] getTrackingIcons(Point p)
+			throws BadLocationException {
 		int offs = textArea.viewToModel(new Point(0, p.y));
 		int line = textArea.getLineOfOffset(offs);
 		return iconArea.getTrackingIcons(line);
 	}
 
 
+	/**
+	 * Returns whether the fold indicator is enabled.
+	 *
+	 * @return Whether the fold indicator is enabled.
+	 * @see #setFoldIndicatorEnabled(boolean)
+	 */
+	public boolean isFoldIndicatorEnabled() {
+		for (int i=0; i<getComponentCount(); i++) {
+			if (getComponent(i)==foldIndicator) {
+				return true;
+			}
+		}
+		return false;
+	}
+
+
 	/**
 	 * Returns whether bookmarking is enabled.
 	 *
@@ -314,6 +459,19 @@ public class Gutter extends JComponent {
 	}
 
 
+	/**
+	 * Sets the color to use to render active line ranges.
+	 *
+	 * @param color The color to use.  If this is null, then the default
+	 *        color is used.
+	 * @see #getActiveLineRangeColor()
+	 * @see #DEFAULT_ACTIVE_LINE_RANGE_COLOR
+	 */
+	public void setActiveLineRangeColor(Color color) {
+		iconArea.setActiveLineRangeColor(color);
+	}
+
+
 	/**
 	 * Highlights a range of lines in the icon area.  This, of course, will
 	 * only be visible if the icon area is visible.
@@ -372,6 +530,7 @@ public class Gutter extends JComponent {
 	/**
 	 * {@inheritDoc}
 	 */
+	@Override
 	public void setComponentOrientation(ComponentOrientation o) {
 		// Reuse the border to preserve its color.
 		if (o.isLeftToRight()) {
@@ -384,6 +543,68 @@ public class Gutter extends JComponent {
 	}
 
 
+	/**
+	 * Sets the icons to use to represent collapsed and expanded folds.
+	 *
+	 * @param collapsedIcon The collapsed fold icon.  This cannot be
+	 *        <code>null</code>.
+	 * @param expandedIcon The expanded fold icon.  This cannot be
+	 *        <code>null</code>.
+	 */
+	public void setFoldIcons(Icon collapsedIcon, Icon expandedIcon) {
+		if (foldIndicator!=null) {
+			foldIndicator.setFoldIcons(collapsedIcon, expandedIcon);
+		}
+	}
+
+
+	/**
+	 * Toggles whether the fold indicator is enabled.
+	 *
+	 * @param enabled Whether the fold indicator should be enabled.
+	 * @see #isFoldIndicatorEnabled()
+	 */
+	public void setFoldIndicatorEnabled(boolean enabled) {
+		if (foldIndicator!=null) {
+			if (enabled) {
+				add(foldIndicator, BorderLayout.LINE_END);
+			}
+			else {
+				remove(foldIndicator);
+			}
+			revalidate();
+		}
+	}
+
+
+	/**
+	 * Sets the background color used by the (default) fold icons.
+	 *
+	 * @param bg The new background color.
+	 * @see #getFoldBackground()
+	 */
+	public void setFoldBackground(Color bg) {
+		if (bg==null) {
+			bg = FoldIndicator.DEFAULT_FOLD_BACKGROUND;
+		}
+		foldIndicator.setFoldIconBackground(bg);
+	}
+
+
+	/**
+	 * Sets the foreground color used by the fold indicator.
+	 *
+	 * @param fg The new fold indicator foreground.
+	 * @see #getFoldIndicatorForeground()
+	 */
+	public void setFoldIndicatorForeground(Color fg) {
+		if (fg==null) {
+			fg = FoldIndicator.DEFAULT_FOREGROUND;
+		}
+		foldIndicator.setForeground(fg);
+	}
+
+
 	/**
 	 * Toggles whether the icon row header (used for breakpoints, bookmarks,
 	 * etc.) is enabled.
@@ -411,7 +632,12 @@ public class Gutter extends JComponent {
 	 * @see #getLineNumberColor()
 	 */
 	public void setLineNumberColor(Color color) {
-		lineNumberList.setForeground(color);
+		if (color!=null && !color.equals(lineNumberColor)) {
+			lineNumberColor = color;
+			if (lineNumberList!=null) {
+				lineNumberList.setForeground(color);
+			}
+		}
 	}
 
 
@@ -425,7 +651,12 @@ public class Gutter extends JComponent {
 		if (font==null) {
 			throw new IllegalArgumentException("font cannot be null");
 		}
-		lineNumberList.setFont(font);
+		if (!font.equals(lineNumberFont)) {
+			lineNumberFont = font;
+			if (lineNumberList!=null) {
+				lineNumberList.setFont(font);
+			}
+		}
 	}
 
 
@@ -438,7 +669,10 @@ public class Gutter extends JComponent {
 	 * @see #getLineNumberingStartIndex()
 	 */
 	public void setLineNumberingStartIndex(int index) {
-		lineNumberList.setLineNumberingStartIndex(index);
+		if (index!=lineNumberingStartIndex) {
+			lineNumberingStartIndex = index;
+			lineNumberList.setLineNumberingStartIndex(index);
+		}
 	}
 
 
@@ -461,6 +695,20 @@ public class Gutter extends JComponent {
 	}
 
 
+	/**
+	 * Toggles whether tool tips should be displayed showing the contents of
+	 * collapsed fold regions when the mouse hovers over a +/- icon.
+	 *
+	 * @param show Whether to show these tool tips.
+	 * @see #getShowCollapsedRegionToolTips()
+	 */
+	public void setShowCollapsedRegionToolTips(boolean show) {
+		if (foldIndicator!=null) {
+			foldIndicator.setShowCollapsedRegionToolTips(show);
+		}
+	}
+
+
 	/**
 	 * Sets the text area being displayed.  This will clear any tracking
 	 * icons currently displayed.
@@ -468,25 +716,45 @@ public class Gutter extends JComponent {
 	 * @param textArea The text area.
 	 */
 	void setTextArea(RTextArea textArea) {
+
 		if (this.textArea!=null) {
 			listener.uninstall();
 		}
-		if (lineNumberList==null) {
-			lineNumberList = new LineNumberList(textArea);
-		}
-		else {
-			lineNumberList.setTextArea(textArea);
-		}
-		if (iconArea==null) {
-			iconArea = new IconRowHeader(textArea);
-		}
-		else {
-			iconArea.setTextArea(textArea);
-		}
+
 		if (textArea!=null) {
+
+			RTextAreaEditorKit kit = (RTextAreaEditorKit)textArea.getUI().
+					getEditorKit(textArea);
+
+			if (lineNumberList==null) {
+				lineNumberList = kit.createLineNumberList(textArea);
+				lineNumberList.setFont(getLineNumberFont());
+				lineNumberList.setForeground(getLineNumberColor());
+				lineNumberList.setLineNumberingStartIndex(
+						getLineNumberingStartIndex());
+			}
+			else {
+				lineNumberList.setTextArea(textArea);
+			}
+			if (iconArea==null) {
+				iconArea = kit.createIconRowHeader(textArea);
+			}
+			else {
+				iconArea.setTextArea(textArea);
+			}
+			if (foldIndicator==null) {
+				foldIndicator = new FoldIndicator(textArea);
+			}
+			else {
+				foldIndicator.setTextArea(textArea);
+			}
+
 			listener.install(textArea);
+
 		}
+
 		this.textArea = textArea;
+
 	}
 
 
@@ -504,31 +772,65 @@ public class Gutter extends JComponent {
 	}
 
 
+//	public void setUI(ComponentUI ui) {
+//
+//		Border gutterBorder = getBorder();
+//
+//		super.setUI(ui);
+//
+//		// Some LaFs, such as WebLookAndFeel, override borders even when
+//		// they aren't UIResources.
+//		Border border = getBorder();
+//		if (border != gutterBorder) {
+//			setBorder(gutterBorder);
+//		}
+//
+//	}
+//
+//
+@Override
+public void setBorder(Border border) {
+	if (border instanceof GutterBorder) {
+		super.setBorder(border);
+	}
+}
 	/**
 	 * The border used by the gutter.
 	 */
 	private static class GutterBorder extends EmptyBorder {
 
 		private Color color;
+		private Rectangle visibleRect;
 
 		public GutterBorder(int top, int left, int bottom, int right) {
 			super(top, left, bottom, right);
 			color = new Color(221, 221, 221);
+			visibleRect = new Rectangle();
 		}
 
 		public Color getColor() {
 			return color;
 		}
 
+		@Override
 		public void paintBorder(Component c, Graphics g, int x, int y,
 								int width, int height) {
+
+			visibleRect = g.getClipBounds(visibleRect);
+			if (visibleRect==null) { // ???
+				visibleRect = ((JComponent)c).getVisibleRect();
+			}
+
 			g.setColor(color);
 			if (left==1) {
-				g.drawLine(0,0, 0,height);
+				g.drawLine(0,visibleRect.y,
+						0,visibleRect.y+visibleRect.height);
 			}
 			else {
-				g.drawLine(width-1,0, width-1,height);
+				g.drawLine(width-1,visibleRect.y,
+						width-1,visibleRect.y+visibleRect.height);
 			}
+
 		}
 
 		public void setColor(Color color) {
@@ -579,6 +881,7 @@ public class Gutter extends JComponent {
 
 		public void changedUpdate(DocumentEvent e) {}
 
+		@Override
 		public void componentResized(java.awt.event.ComponentEvent e) {
 			revalidate();
 		}
@@ -603,7 +906,9 @@ public class Gutter extends JComponent {
 			textArea.getDocument().addDocumentListener(this);
 			textArea.addPropertyChangeListener(this);
 			if (textArea instanceof RSyntaxTextArea) {
-				((RSyntaxTextArea)textArea).addActiveLineRangeListener(this);
+				RSyntaxTextArea rsta = (RSyntaxTextArea)textArea;
+				rsta.addActiveLineRangeListener(this);
+				rsta.getFoldManager().addPropertyChangeListener(this);
 			}
 			installed = true;
 		}
@@ -623,6 +928,22 @@ public class Gutter extends JComponent {
 				}
 			}
 
+			// If they toggle whether code folding is enabled...
+			else if (RSyntaxTextArea.CODE_FOLDING_PROPERTY.equals(name)) {
+				boolean foldingEnabled = ((Boolean)e.getNewValue()).
+															booleanValue();
+				if (lineNumberList!=null) { // Its size depends on folding
+					//lineNumberList.revalidate();
+					lineNumberList.updateCellWidths();
+				}
+				setFoldIndicatorEnabled(foldingEnabled);
+			}
+
+			// If code folds are updated...
+			else if (FoldManager.PROPERTY_FOLDS_UPDATED.equals(name)) {
+				repaint();
+			}
+
 		}
 
 		public void removeUpdate(DocumentEvent e) {
@@ -634,7 +955,9 @@ public class Gutter extends JComponent {
 				textArea.removeComponentListener(this);
 				textArea.getDocument().removeDocumentListener(this);
 				if (textArea instanceof RSyntaxTextArea) {
-					((RSyntaxTextArea)textArea).removeActiveLineRangeListener(this);
+					RSyntaxTextArea rsta = (RSyntaxTextArea)textArea;
+					rsta.removeActiveLineRangeListener(this);
+					rsta.getFoldManager().removePropertyChangeListener(this);
 				}
 				installed = false;
 			}
