@@ -2,23 +2,9 @@
  * 02/19/2009
  *
  * GutterIconInfo.java - Information about an Icon in a Gutter.
- * Copyright (C) 2009 Robert Futrell
- * robert_futrell at users.sourceforge.net
- * http://fifesoft.com/rsyntaxtextarea
- *
- * This library is free software; you can redistribute it and/or
- * modify it under the terms of the GNU Lesser General Public
- * License as published by the Free Software Foundation; either
- * version 2.1 of the License, or (at your option) any later version.
- *
- * This library is distributed in the hope that it will be useful,
- * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
- * Lesser General Public License for more details.
- *
- * You should have received a copy of the GNU Lesser General Public
- * License along with this library; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA.
+ * 
+ * This library is distributed under a modified BSD license.  See the included
+ * RSyntaxTextArea.License.txt file for details.
  */
 package org.fife.ui.rtextarea;
 
@@ -56,4 +42,12 @@ public interface GutterIconInfo {
 	 */
 	public int getMarkedOffset();
 
+
+	/**
+	 * Returns the tool tip to display when the mouse hovers over this icon.
+	 *
+	 * @return The tool tip to display.
+	 */
+	public String getToolTip();
+
 }
\ No newline at end of file
