@@ -2,23 +2,9 @@
  * 11/14/2003
  *
  * RTextArea.java - An extension of JTextArea that adds many features.
- * Copyright (C) 2003 Robert Futrell
- * robert_futrell at users.sourceforge.net
- * http://fifesoft.com/rsyntaxtextarea
- *
- * This library is free software; you can redistribute it and/or
- * modify it under the terms of the GNU Lesser General Public
- * License as published by the Free Software Foundation; either
- * version 2.1 of the License, or (at your option) any later version.
- *
- * This library is distributed in the hope that it will be useful,
- * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
- * Lesser General Public License for more details.
- *
- * You should have received a copy of the GNU Lesser General Public
- * License along with this library; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA.
+ * 
+ * This library is distributed under a modified BSD license.  See the included
+ * RSyntaxTextArea.License.txt file for details.
  */
 package org.fife.ui.rtextarea;
 
@@ -36,7 +22,6 @@ import java.io.IOException;
 import java.io.ObjectInputStream;
 import java.io.ObjectOutputStream;
 import java.io.Reader;
-import java.io.Serializable;
 import java.util.ArrayList;
 import java.util.List;
 import java.util.Locale;
@@ -58,6 +43,7 @@ import javax.swing.text.DefaultEditorKit;
 import javax.swing.text.Document;
 import javax.swing.text.Element;
 import javax.swing.text.Highlighter;
+import javax.swing.text.Segment;
 import javax.swing.undo.CannotRedoException;
 import javax.swing.undo.CannotUndoException;
 
@@ -90,8 +76,7 @@ import org.fife.ui.rtextarea.Macro.MacroRecord;
  * @author Robert Futrell
  * @version 1.0
  */
-public class RTextArea extends RTextAreaBase
-								implements Printable, Serializable {
+public class RTextArea extends RTextAreaBase implements Printable {
 
 	/**
 	 * Constant representing insert mode.
@@ -141,11 +126,22 @@ public class RTextArea extends RTextAreaBase
 	 */
 	private JPopupMenu popupMenu;
 
+	private JMenuItem undoMenuItem;
+	private JMenuItem redoMenuItem;
+	private JMenuItem cutMenuItem;
+	private JMenuItem pasteMenuItem;
+	private JMenuItem deleteMenuItem;
+
 	/**
 	 * Whether the popup menu has been created.
 	 */
 	private boolean popupMenuCreated;
 
+	/**
+	 * The text last searched for via Ctrl+K or Ctrl+Shift+K.
+	 */
+	private static String selectedOccurrenceText;
+
 	/**
 	 * Can return tool tips for this text area.  Subclasses can install a
 	 * supplier as a means of adding custom tool tips without subclassing
@@ -168,7 +164,7 @@ public class RTextArea extends RTextAreaBase
 
 	private transient LineHighlightManager lineHighlightManager;
 
-	private ArrayList markAllHighlights;		// Highlights from "mark all".
+	private ArrayList<Object> markAllHighlights;		// Highlights from "mark all".
 	private String markedWord;				// Expression marked in "mark all."
 	private ChangeableHighlightPainter markAllHighlightPainter;
 
@@ -389,21 +385,51 @@ public class RTextArea extends RTextAreaBase
 	 * custom configuration (configuring actions, adding/removing items, etc.).
 	 * <p>
 	 *
-	 * The default implementation does nothing.
+	 * The default implementation does nothing.<p>
+	 * 
+	 * If you set the popup menu via {@link #setPopupMenu(JPopupMenu)}, you
+	 * will want to override this method, especially if you removed any of the
+	 * menu items in the default popup menu.
 	 *
 	 * @param popupMenu The popup menu.  This will never be <code>null</code>.
 	 * @see #createPopupMenu()
 	 * @see #setPopupMenu(JPopupMenu)
 	 */
 	protected void configurePopupMenu(JPopupMenu popupMenu) {
+
+		boolean canType = isEditable() && isEnabled();
+
+		// Since the user can customize the popup menu, these actions may not
+		// have been created.
+		if (undoMenuItem!=null) {
+			undoMenuItem.setEnabled(undoAction.isEnabled() && canType);
+			redoMenuItem.setEnabled(redoAction.isEnabled() && canType);
+			cutMenuItem.setEnabled(cutAction.isEnabled() && canType);
+			pasteMenuItem.setEnabled(pasteAction.isEnabled() && canType);
+			deleteMenuItem.setEnabled(deleteAction.isEnabled() && canType);
+		}
+
 	}
 
 
+	/**
+	 * Creates the default implementation of the model to be used at
+	 * construction if one isn't explicitly given. A new instance of RDocument
+	 * is returned.
+	 *
+	 * @return The default document.
+	 */
+	@Override
+	protected Document createDefaultModel() {
+		return new RDocument();
+	}
+
 	/**
 	 * Returns the caret event/mouse listener for <code>RTextArea</code>s.
 	 *
 	 * @return The caret event/mouse listener.
 	 */
+	@Override
 	protected RTAMouseListener createMouseListener() {
 		return new RTextAreaMutableCaretEvent(this);
 	}
@@ -416,110 +442,79 @@ public class RTextArea extends RTextAreaBase
 	 * @return The popup menu.
 	 * @see #setPopupMenu(JPopupMenu)
 	 * @see #configurePopupMenu(JPopupMenu)
+	 * @see #createPopupMenuItem(Action)
 	 */
 	protected JPopupMenu createPopupMenu() {
-
 		JPopupMenu menu = new JPopupMenu();
-		JMenuItem menuItem;
-
-		menuItem = new JMenuItem(undoAction);
-		menuItem.setAccelerator(null);
-		menuItem.setToolTipText(null);
-		menu.add(menuItem);
-
-		menuItem = new JMenuItem(redoAction);
-		menuItem.setAccelerator(null);
-		menuItem.setToolTipText(null);
-		menu.add(menuItem);
-
+		menu.add(undoMenuItem = createPopupMenuItem(undoAction));
+		menu.add(redoMenuItem = createPopupMenuItem(redoAction));
 		menu.addSeparator();
-
-		menuItem = new JMenuItem(cutAction);
-		menuItem.setAccelerator(null);
-		menuItem.setToolTipText(null);
-		menu.add(menuItem);
-
-		menuItem = new JMenuItem(copyAction);
-		menuItem.setAccelerator(null);
-		menuItem.setToolTipText(null);
-		menu.add(menuItem);
-
-		menuItem = new JMenuItem(pasteAction);
-		menuItem.setAccelerator(null);
-		menuItem.setToolTipText(null);
-		menu.add(menuItem);
-
-		menuItem = new JMenuItem(deleteAction);
-		menuItem.setAccelerator(null);
-		menuItem.setToolTipText(null);
-		menu.add(menuItem);
-
+		menu.add(cutMenuItem = createPopupMenuItem(cutAction));
+		menu.add(createPopupMenuItem(copyAction));
+		menu.add(pasteMenuItem = createPopupMenuItem(pasteAction));
+		menu.add(deleteMenuItem = createPopupMenuItem(deleteAction));
 		menu.addSeparator();
-
-		menuItem = new JMenuItem(selectAllAction);
-		menuItem.setAccelerator(null);
-		menuItem.setToolTipText(null);
-		menu.add(menuItem);
-
+		menu.add(createPopupMenuItem(selectAllAction));
 		return menu;
-
 	}
 
 
 	/**
 	 * Creates the actions used in the popup menu and retrievable by
 	 * {@link #getAction(int)}.
+	 * TODO: Remove these horrible hacks and move localizing of actions into
+	 * the editor kits, where it should be!  The context menu should contain
+	 * actions from the editor kits.
 	 */
 	private static void createPopupMenuActions() {
 
-		ResourceBundle bundle = ResourceBundle.getBundle(MSG);
-
 		// Create actions for right-click popup menu.
 		// 1.5.2004/pwy: Replaced the CTRL_MASK with the cross-platform version...
 		int mod = Toolkit.getDefaultToolkit().getMenuShortcutKeyMask();
+		ResourceBundle msg = ResourceBundle.getBundle(MSG);
 
-		String name = bundle.getString("CutName");
-		char mnemonic = bundle.getString("CutMnemonic").charAt(0);
-		String desc = bundle.getString("CutDesc");
-		cutAction = new RTextAreaEditorKit.CutAction(name, null, desc,
-			new Integer(mnemonic), KeyStroke.getKeyStroke(KeyEvent.VK_X, mod));
+		cutAction = new RTextAreaEditorKit.CutAction();
+		cutAction.setProperties(msg, "Action.Cut");
+		cutAction.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_X, mod));
+		copyAction = new RTextAreaEditorKit.CopyAction();
+		copyAction.setProperties(msg, "Action.Copy");
+		copyAction.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_C, mod));
+		pasteAction = new RTextAreaEditorKit.PasteAction();
+		pasteAction.setProperties(msg, "Action.Paste");
+		pasteAction.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_V, mod));
+		deleteAction = new RTextAreaEditorKit.DeleteNextCharAction();
+		deleteAction.setProperties(msg, "Action.Delete");
+		deleteAction.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_DELETE, 0));
+		undoAction = new RTextAreaEditorKit.UndoAction();
+		undoAction.setProperties(msg, "Action.Undo");
+		undoAction.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_Z, mod));
+		redoAction = new RTextAreaEditorKit.RedoAction();
+		redoAction.setProperties(msg, "Action.Redo");
+		redoAction.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_Y, mod));
+		selectAllAction = new RTextAreaEditorKit.SelectAllAction();
+		selectAllAction.setProperties(msg, "Action.SelectAll");
+		selectAllAction.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_A, mod));
 
-		name = bundle.getString("CopyName");
-		mnemonic = bundle.getString("CopyMnemonic").charAt(0);
-		desc = bundle.getString("CopyDesc");
-		copyAction = new RTextAreaEditorKit.CopyAction(name, null, desc,
-			new Integer(mnemonic), KeyStroke.getKeyStroke(KeyEvent.VK_C, mod));
+	}
 
-		name = bundle.getString("PasteName");
-		mnemonic = bundle.getString("PasteMnemonic").charAt(0);
-		desc = bundle.getString("PasteDesc");
-		pasteAction = new RTextAreaEditorKit.PasteAction(name, null, desc,
-			new Integer(mnemonic), KeyStroke.getKeyStroke(KeyEvent.VK_V, mod));
-
-		name = bundle.getString("DeleteName");
-		mnemonic = bundle.getString("DeleteMnemonic").charAt(0);
-		desc = bundle.getString("DeleteDesc");
-		deleteAction = new RTextAreaEditorKit.DeleteNextCharAction(name, null, desc,
-			new Integer(mnemonic), KeyStroke.getKeyStroke(KeyEvent.VK_DELETE, 0));
-
-		name = bundle.getString("CantUndoName");
-		mnemonic = bundle.getString("UndoMnemonic").charAt(0);
-		desc = bundle.getString("UndoDesc");
-		undoAction = new RTextAreaEditorKit.UndoAction(name, null, desc,
-			new Integer(mnemonic), KeyStroke.getKeyStroke(KeyEvent.VK_Z, mod));
-
-		name = bundle.getString("CantRedoName");
-		mnemonic = bundle.getString("RedoMnemonic").charAt(0);
-		desc = bundle.getString("RedoDesc");
-		redoAction = new RTextAreaEditorKit.RedoAction(name, null, desc,
-			new Integer(mnemonic), KeyStroke.getKeyStroke(KeyEvent.VK_Y, mod));
-
-		name = bundle.getString("SAName");
-		mnemonic = bundle.getString("SAMnemonic").charAt(0);
-		desc = bundle.getString("SelectAllDesc");
-		selectAllAction = new RTextAreaEditorKit.SelectAllAction(name, null, desc,
-			new Integer(mnemonic), KeyStroke.getKeyStroke(KeyEvent.VK_A, mod));
 
+	/**
+	 * Creates and configures a menu item for used in the popup menu.
+	 *
+	 * @param a The action for the menu item.
+	 * @return The menu item.
+	 * @see #createPopupMenu()
+	 */
+	protected JMenuItem createPopupMenuItem(Action a) {
+		JMenuItem item = new JMenuItem(a) {
+			@Override
+			public void setToolTipText(String text) {
+				// Ignore!  Actions (e.g. undo/redo) set this when changing
+				// their text due to changing enabled state.
+			}
+		};
+		item.setAccelerator(null);
+		return item;
 	}
 
 
@@ -528,11 +523,37 @@ public class RTextArea extends RTextAreaBase
 	 *
 	 * @return The UI.
 	 */
+	@Override
 	protected RTextAreaUI createRTextAreaUI() {
 		return new RTextAreaUI(this);
 	}
 
 
+	/**
+	 * Creates a string of space characters of the specified size.
+	 *
+	 * @param size The number of spaces.
+	 * @return The string of spaces.
+	 */
+	private final String createSpacer(int size) {
+		StringBuilder sb = new StringBuilder();
+		for (int i=0; i<size; i++) {
+			sb.append(' ');
+		}
+		return sb.toString();
+	}
+
+
+	/**
+	 * Creates an undo manager for use in this text area.
+	 *
+	 * @return The undo manager.
+	 */
+	protected RUndoManager createUndoManager() {
+		return new RUndoManager(this);
+	}
+
+
 	/**
 	 * Removes all undoable edits from this document's undo manager.  This
 	 * method also makes the undo/redo actions disabled.
@@ -546,7 +567,7 @@ public class RTextArea extends RTextAreaBase
 	public void discardAllEdits() {
 		undoManager.discardAllEdits();
 		getDocument().removeUndoableEditListener(undoManager);
-		undoManager = new RUndoManager(this);
+		undoManager = createUndoManager();
 		getDocument().addUndoableEditListener(undoManager);
 		undoManager.updateActions();
 	}
@@ -586,6 +607,7 @@ public class RTextArea extends RTextAreaBase
 	 *
 	 * @param e The caret event.
 	 */
+	@Override
 	protected void fireCaretUpdate(CaretEvent e) {
 
 		// Decide whether we need to repaint the current line background.
@@ -763,6 +785,17 @@ public class RTextArea extends RTextAreaBase
 	}
 
 
+	/**
+	 * Returns the text last selected and used in a Ctrl+K operation.
+	 *
+	 * @return The text, or <code>null</code> if none.
+	 * @see #setSelectedOccurrenceText(String)
+	 */
+	public static String getSelectedOccurrenceText() {
+		return selectedOccurrenceText;
+	}
+
+
 	/**
 	 * Returns the text mode this editor pane is currently in.
 	 *
@@ -798,6 +831,7 @@ public class RTextArea extends RTextAreaBase
 	 * @see #getToolTipSupplier()
 	 * @see #setToolTipSupplier(ToolTipSupplier)
 	 */
+	@Override
 	public String getToolTipText(MouseEvent e) {
 		String tip = null;
 		if (getToolTipSupplier()!=null) {
@@ -815,7 +849,7 @@ public class RTextArea extends RTextAreaBase
 	 * @param content The content to add.
 	 */
 	protected void handleReplaceSelection(String content) {
-		// Call into super to handle composed text (1.5+ only though).
+		// Call into super to handle composed text.
 		super.replaceSelection(content);
 	}
 
@@ -838,7 +872,7 @@ public class RTextArea extends RTextAreaBase
 		}
 
 		// Install the undo manager.
-		undoManager = new RUndoManager(this);
+		undoManager = createUndoManager();
 		getDocument().addUndoableEditListener(undoManager);
 
 		// Set the defaults for various stuff.
@@ -905,12 +939,17 @@ public class RTextArea extends RTextAreaBase
 			if (markAllHighlights!=null)
 				clearMarkAllHighlights();
 			else
-				markAllHighlights = new ArrayList(10);
+				markAllHighlights = new ArrayList<Object>();
 			int caretPos = getCaretPosition();
 			markedWord = toMark;
 			setCaretPosition(0);
-			boolean found = SearchEngine.find(this, toMark, true, matchCase,
-										wholeWord, regex);
+			SearchContext context = new SearchContext();
+			context.setSearchFor(toMark);
+			context.setMatchCase(matchCase);
+			context.setRegularExpression(regex);
+			context.setSearchForward(true);
+			context.setWholeWord(wholeWord);
+			boolean found = SearchEngine.find(this, context);
 			while (found) {
 				int start = getSelectionStart();
 				int end = getSelectionEnd();
@@ -921,8 +960,7 @@ public class RTextArea extends RTextAreaBase
 					ble.printStackTrace();
 				}
 				numMarked++;
-				found = SearchEngine.find(this, toMark, true,
-									matchCase, wholeWord, regex);
+				found = SearchEngine.find(this, context);
 			}
 			setCaretPosition(caretPos);
 			repaint();
@@ -934,6 +972,7 @@ public class RTextArea extends RTextAreaBase
 	/**
 	 * {@inheritDoc}
 	 */
+	@Override
 	public void paste() {
 		// Treat paste operations as atomic, otherwise the removal and
 		// insertion are treated as two separate undo-able operations.
@@ -951,20 +990,17 @@ public class RTextArea extends RTextAreaBase
 	 */
 	public synchronized void playbackLastMacro() {
 		if (currentMacro!=null) {
-			Action[] actions = getActions();
-			int numActions = actions.length;
-			List macroRecords = currentMacro.getMacroRecords();
-			int num = macroRecords.size();
-			if (num>0) {
+			List<MacroRecord> macroRecords = currentMacro.getMacroRecords();
+			if (!macroRecords.isEmpty()) {
+				Action[] actions = getActions();
 				undoManager.beginInternalAtomicEdit();
 				try {
-					for (int i=0; i<num; i++) {
-						MacroRecord record = (MacroRecord)macroRecords.get(i);
-						for (int j=0; j<numActions; j++) {
-							if ((actions[j] instanceof RecordableTextAction) &&
+					for (MacroRecord record : macroRecords) {
+						for (int i=0; i<actions.length; i++) {
+							if ((actions[i] instanceof RecordableTextAction) &&
 								record.id.equals(
-								((RecordableTextAction)actions[j]).getMacroID())) {
-								actions[j].actionPerformed(
+								((RecordableTextAction)actions[i]).getMacroID())) {
+								actions[i].actionPerformed(
 									new ActionEvent(this,
 												ActionEvent.ACTION_PERFORMED,
 												record.actionCommand));
@@ -998,6 +1034,7 @@ public class RTextArea extends RTextAreaBase
 	 * new <code>Document</code>, thus requiring us to re-attach our Undo
 	 * manager.  With this version we just replace the text.
 	 */
+	@Override
 	public void read(Reader in, Object desc) throws IOException {
 
 		RTextAreaEditorKit kit = (RTextAreaEditorKit)getUI().getEditorKit(this);
@@ -1029,7 +1066,7 @@ public class RTextArea extends RTextAreaBase
 
 		// UndoManagers cannot be serialized without Exceptions.  See
 		// http://bugs.sun.com/bugdatabase/view_bug.do?bug_id=4275892
-		undoManager = new RUndoManager(this);
+		undoManager = createUndoManager();
 		getDocument().addUndoableEditListener(undoManager);
 
 		lineHighlightManager = null; // Keep FindBugs happy.
@@ -1098,6 +1135,7 @@ public class RTextArea extends RTextAreaBase
 	 * @see #insert(String, int)
 	 * @see #replaceRange(String, int, int)
 	 */
+	@Override
 	public void replaceRange(String str, int start, int end) {
 		if (end < start)
 			throw new IllegalArgumentException("end before start");
@@ -1130,6 +1168,7 @@ public class RTextArea extends RTextAreaBase
 	 *
 	 * @param text The content to replace the selection with.
 	 */
+	@Override
 	public void replaceSelection(String text) {
 
 		// It's legal for null to be used here...
@@ -1138,8 +1177,16 @@ public class RTextArea extends RTextAreaBase
 			return;
 		}
 
-		if (getTabsEmulated() && text.indexOf('\t')>-1) {
-			text = replaceTabsWithSpaces(text);
+		if (getTabsEmulated()) {
+			int firstTab = text.indexOf('\t');
+			if (firstTab>-1) {
+				int docOffs = getSelectionStart();
+				try {
+					text = replaceTabsWithSpaces(text, docOffs, firstTab);
+				} catch (BadLocationException ble) { // Never happens
+					ble.printStackTrace();
+				}
+			}
 		}
 
 		// If the user wants to overwrite text...
@@ -1181,56 +1228,89 @@ public class RTextArea extends RTextAreaBase
 	}
 
 
-	private StringBuffer repTabsSB;
+	private static StringBuilder repTabsSB;
+	private static Segment repTabsSeg = new Segment();
 	/**
 	 * Replaces all instances of the tab character in <code>text</code> with
 	 * the number of spaces equivalent to a tab in this text area.<p>
 	 *
 	 * This method should only be called from thread-safe methods, such as
-	 * {@link replaceSelection(String)}.
+	 * {@link #replaceSelection(String)}.
 	 *
 	 * @param text The <code>java.lang.String</code> in which to replace tabs
 	 *        with spaces.  This has already been verified to have at least
 	 *        one tab character in it.
-	 * @return A <code>java.lang.String</code> just like <code>text</code>,
-	 *         but with spaces instead of tabs.
+	 * @param docOffs The offset in the document at which the text is being
+	 *        inserted.
+	 * @param firstTab The offset into <code>text</code> of the first tab.  Assumed
+	 *        to be &gt;= 0.
+	 * @return A <code>String</code> just like <code>text</code>, but with
+	 *         spaces instead of tabs.
 	 */
-	private final String replaceTabsWithSpaces(final String text) {
+	private final String replaceTabsWithSpaces(String text, int docOffs, int firstTab)
+			throws BadLocationException {
 
-		String tabText = "";
-		int temp = getTabSize();
-		for (int i=0; i<temp; i++) {
-			tabText += ' ';
+		int tabSize = getTabSize();
+
+		// Get how many chars into the current line we are
+		Document doc = getDocument();
+		Element root = doc.getDefaultRootElement();
+		int lineIndex = root.getElementIndex(docOffs);
+		Element line = root.getElement(lineIndex);
+		int lineStart = line.getStartOffset();
+		int charCount = docOffs - lineStart;
+
+		// Figure out how many chars into the "current tab" we are
+		if (charCount>0) {
+			doc.getText(lineStart, charCount, repTabsSeg);
+			charCount = 0;
+			for (int i=0; i<repTabsSeg.count; i++) {
+				char ch = repTabsSeg.array[repTabsSeg.offset + i];
+				if (ch=='\t') {
+					charCount = 0;
+				}
+				else {
+					charCount = (charCount + 1) % tabSize;
+				}
+			}
 		}
 
-		// Common case: User's entering a single tab (pressed the tab key).
+		// Common case: The user's entering a single tab (pressed the tab key).
 		if (text.length()==1) {
-			return tabText;
+			return createSpacer(tabSize - charCount);
 		}
 
-		// Otherwise, there may be more than one tab.  Manually search for
-		// tabs for performance, as opposed to using String#replaceAll().
-		// This method is called for each character inserted when "replace
-		// tabs with spaces" is enabled, so we need to be quick.
+		// Otherwise, there may be more than one tab.
 
-		//return text.replaceAll("\t", tabText);
 		if (repTabsSB==null) {
-			repTabsSB = new StringBuffer();
+			repTabsSB = new StringBuilder();
 		}
 		repTabsSB.setLength(0);
-		char[] array = text.toCharArray(); // Wouldn't be needed in 1.5!
-		int oldPos = 0;
-		int pos = 0;
-		while ((pos=text.indexOf('\t', oldPos))>-1) {
-			//repTabsSB.append(text, oldPos, pos); // Added in Java 1.5
-			if (pos>oldPos) {
-				repTabsSB.append(array, oldPos, pos-oldPos);
+		char[] array = text.toCharArray();
+		int lastPos = 0;
+		int offsInLine = charCount; // Accurate enough for our start
+		for (int pos=firstTab; pos<array.length; pos++) {
+			char ch = array[pos];
+			switch (ch) {
+				case '\t':
+					if (pos>lastPos) {
+						repTabsSB.append(array, lastPos, pos-lastPos);
+					}
+					int thisTabSize = tabSize - (offsInLine%tabSize);
+					repTabsSB.append(createSpacer(thisTabSize));
+					lastPos = pos + 1;
+					offsInLine = 0;
+					break;
+				case '\n':
+					offsInLine = 0;
+					break;
+				default:
+					offsInLine++;
+					break;
 			}
-			repTabsSB.append(tabText);
-			oldPos = pos + 1;
 		}
-		if (oldPos<array.length) {
-			repTabsSB.append(array, oldPos, array.length-oldPos);
+		if (lastPos<array.length) {
+			repTabsSB.append(array, lastPos, array.length-lastPos);
 		}
 
 		return repTabsSB.toString();
@@ -1249,7 +1329,7 @@ public class RTextArea extends RTextAreaBase
 	 */
 	public static void setActionProperties(int action, String name,
 							char mnemonic, KeyStroke accelerator) {
-		setActionProperties(action, name, new Integer(mnemonic), accelerator);
+		setActionProperties(action, name, Integer.valueOf(mnemonic), accelerator);
 	}
 
 
@@ -1297,10 +1377,10 @@ public class RTextArea extends RTextAreaBase
 
 
 	/**
-	 * This method is overridden to make sure that instances of
-	 * <code>RTextArea</code> only use {@link ConfigurableCaret}s.
-	 * To set the style of caret (vertical line, block, etc.) used for
-	 * insert or overwrite mode, use {@link #setCaretStyle(int, int)}.
+	 * Sets the caret to use in this text area.  It is strongly encouraged to
+	 * use {@link ConfigurableCaret}s (which is used by default), or a
+	 * subclass, since they know how to render themselves differently when the
+	 * user toggles between insert and overwrite modes.
 	 *
 	 * @param caret The caret to use.  If this is not an instance of
 	 *        <code>ConfigurableCaret</code>, an exception is thrown.
@@ -1308,13 +1388,11 @@ public class RTextArea extends RTextAreaBase
 	 *         <code>ConfigurableCaret</code>.
 	 * @see #setCaretStyle(int, int)
 	 */
+	@Override
 	public void setCaret(Caret caret) {
-		if (!(caret instanceof ConfigurableCaret)) {
-			throw new IllegalArgumentException(
-						"RTextArea needs ConfigurableCaret");
-		}
 		super.setCaret(caret);
-		if (carets!=null) { // Called by setUI() before carets is initialized
+		if (carets!=null && // Called by setUI() before carets is initialized
+				caret instanceof ConfigurableCaret) {
 			((ConfigurableCaret)caret).setStyle(carets[getTextMode()]);
 		}
 	}
@@ -1333,7 +1411,7 @@ public class RTextArea extends RTextAreaBase
 					style<=ConfigurableCaret.MAX_STYLE ?
 						style : ConfigurableCaret.THICK_VERTICAL_LINE_STYLE);
 		carets[mode] = style;
-		if (mode==getTextMode()) {
+		if (mode==getTextMode() && getCaret() instanceof ConfigurableCaret) {
 			// Will repaint the caret if necessary.
 			((ConfigurableCaret)getCaret()).setStyle(style);
 		}
@@ -1345,12 +1423,13 @@ public class RTextArea extends RTextAreaBase
 	 *
 	 * @param document The new document to use.
 	 * @throws IllegalArgumentException If the document is not an instance of
-	 *         {@link AbstractDocument}.
+	 *         {@link RDocument}.
 	 */
+	@Override
 	public void setDocument(Document document) {
-		if (!(document instanceof AbstractDocument)) {
+		if (!(document instanceof RDocument)) {
 			throw new IllegalArgumentException("RTextArea requires " +
-				"instances of AbstractDocument for its document");
+				"instances of RDocument for its document");
 		}
 		if (undoManager!=null) { // First time through, undoManager==null
 			Document old = getDocument();
@@ -1423,11 +1502,16 @@ public class RTextArea extends RTextAreaBase
 
 
 	/**
-	 * Sets the popup menu used by this text area.
+	 * Sets the popup menu used by this text area.<p>
+	 * 
+	 * If you set the popup menu with this method, you'll want to consider also
+	 * overriding {@link #configurePopupMenu(JPopupMenu)}, especially if you
+	 * removed any of the default menu items.
 	 *
 	 * @param popupMenu The popup menu.  If this is <code>null</code>, no
 	 *        popup menu will be displayed.
 	 * @see #getPopupMenu()
+	 * @see #configurePopupMenu(JPopupMenu)
 	 */
 	public void setPopupMenu(JPopupMenu popupMenu) {
 		this.popupMenu = popupMenu;
@@ -1438,6 +1522,7 @@ public class RTextArea extends RTextAreaBase
 	/**
 	 * {@inheritDoc}
 	 */
+	@Override
 	public void setRoundedSelectionEdges(boolean rounded) {
 		if (getRoundedSelectionEdges()!=rounded) {
 			markAllHighlightPainter.setRoundedEdges(rounded);
@@ -1447,9 +1532,31 @@ public class RTextArea extends RTextAreaBase
 
 
 	/**
-	 * Sets the text mode for this editor pane.
+	 * Sets the text last selected/Ctrl+K'd in an <code>RTextArea</code>.
+	 * This text will be searched for in subsequent Ctrl+K/Ctrl+Shift+K
+	 * actions (Cmd+K on OS X).<p>
+	 *
+	 * Since the selected occurrence actions are built into RTextArea,
+	 * applications usually do not have to call this method directly, but can
+	 * choose to do so if they wish (for example, if they wish to set this
+	 * value when the user does a search via a Find dialog).
+	 *
+	 * @param text The selected text.
+	 * @see #getSelectedOccurrenceText()
+	 */
+	public static void setSelectedOccurrenceText(String text) {
+		selectedOccurrenceText = text;
+	}
+
+
+	/**
+	 * Sets the text mode for this editor pane.  If the currently installed
+	 * caret is an instance of {@link ConfigurableCaret}, it will be
+	 * automatically updated to render itself appropriately for the new text
+	 * mode.
 	 *
 	 * @param mode Either {@link #INSERT_MODE} or {@link #OVERWRITE_MODE}.
+	 * @see #getTextMode()
 	 */
 	public void setTextMode(int mode) {
 
@@ -1457,8 +1564,10 @@ public class RTextArea extends RTextAreaBase
 			mode = INSERT_MODE;
 
 		if (textMode != mode) {
-			ConfigurableCaret cc = (ConfigurableCaret)getCaret();
-			cc.setStyle(carets[mode]);
+			Caret caret = getCaret();
+			if (caret instanceof ConfigurableCaret) {
+				((ConfigurableCaret)caret).setStyle(carets[mode]);
+			}
 			textMode = mode;
 		}
 
@@ -1487,6 +1596,7 @@ public class RTextArea extends RTextAreaBase
 	 *
 	 * @param ui This parameter is ignored.
 	 */
+	@Override
 	public final void setUI(TextUI ui) {
 
 		// Update the popup menu's ui.
@@ -1549,6 +1659,7 @@ public class RTextArea extends RTextAreaBase
 			super(textArea);
 		}
 
+		@Override
 		public void focusGained(FocusEvent e) {
 			Caret c = getCaret();
 			boolean enabled = c.getDot()!=c.getMark();
@@ -1557,9 +1668,11 @@ public class RTextArea extends RTextAreaBase
 			undoManager.updateActions(); // To reflect this text area.
 		}
 
+		@Override
 		public void focusLost(FocusEvent e) {
 		}
 
+		@Override
 		public void mouseDragged(MouseEvent e) {
 			if ((e.getModifiers() & MouseEvent.BUTTON1_MASK) != 0) {
 				Caret caret = getCaret();
@@ -1569,6 +1682,7 @@ public class RTextArea extends RTextAreaBase
 			}
 		}
 
+		@Override
 		public void mousePressed(MouseEvent e) {
 			// WORKAROUND:  Since JTextComponent only updates the caret
 			// location on mouse clicked and released, we'll do it on dragged
@@ -1581,6 +1695,7 @@ public class RTextArea extends RTextAreaBase
 			}
 		}
 
+		@Override
 		public void mouseReleased(MouseEvent e) {
 			if ((e.getModifiers()&MouseEvent.BUTTON3_MASK)!=0)
 				showPopup(e);
