@@ -2,23 +2,9 @@
  * 02/19/2006
  *
  * SearchEngine.java - Handles find/replace operations in an RTextArea.
- * Copyright (C) 2006 Robert Futrell
- * robert_futrell at users.sourceforge.net
- * http://fifesoft.com/rsyntaxtextarea
- *
- * This library is free software; you can redistribute it and/or
- * modify it under the terms of the GNU Lesser General Public
- * License as published by the Free Software Foundation; either
- * version 2.1 of the License, or (at your option) any later version.
- *
- * This library is distributed in the hope that it will be useful,
- * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
- * Lesser General Public License for more details.
- *
- * You should have received a copy of the GNU Lesser General Public
- * License along with this library; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA.
+ * 
+ * This library is distributed under a modified BSD license.  See the included
+ * RSyntaxTextArea.License.txt file for details.
  */
 package org.fife.ui.rtextarea;
 
@@ -34,13 +20,24 @@ import javax.swing.JTextArea;
 import javax.swing.text.BadLocationException;
 import javax.swing.text.Caret;
 
+import org.fife.ui.rsyntaxtextarea.RSyntaxTextArea;
+import org.fife.ui.rsyntaxtextarea.folding.FoldManager;
+
 
 /**
  * A singleton class that can perform advanced find/replace operations
- * in an {@link RTextArea}.
+ * in an {@link RTextArea}.  Simply create a {@link SearchContext} and call
+ * one of the following methods:
+ * 
+ * <ul>
+ *    <li>{@link #find(JTextArea, SearchContext)}
+ *    <li>{@link #replace(RTextArea, SearchContext)}
+ *    <li>{@link #replaceAll(RTextArea, SearchContext)}
+ * </ul>
  *
  * @author Robert Futrell
  * @version 1.0
+ * @see SearchContext
  */
 public class SearchEngine {
 
@@ -58,31 +55,26 @@ public class SearchEngine {
 	 * text area.
 	 *
 	 * @param textArea The text area in which to search.
-	 * @param text The string literal or regular expression to search for.
-	 * @param forward Whether to search forward from the caret position or
-	 *        backward from it.
-	 * @param matchCase Whether the search should be case-sensitive.
-	 * @param wholeWord Whether there should be spaces or tabs on either side
-	 *        of the match.
-	 * @param regex Whether <code>text</code> is a Java regular expression to
-	 *        search for.
+	 * @param context What to search for and all search options.
 	 * @return Whether a match was found (and thus selected).
-	 * @throws PatternSyntaxException If <code>regex</code> is
-	 *         <code>true</code> but <code>text</code> is not a valid regular
-	 *         expression.
-	 * @see #replace
-	 * @see #regexReplace
+	 * @throws PatternSyntaxException If this is a regular expression search
+	 *         but the search text is an invalid regular expression.
+	 * @see #replace(RTextArea, SearchContext)
+	 * @see #replaceAll(RTextArea, SearchContext)
 	 */
-	public static boolean find(JTextArea textArea, String text,
-							boolean forward, boolean matchCase,
-							boolean wholeWord, boolean regex)
-									throws PatternSyntaxException {
+	public static boolean find(JTextArea textArea, SearchContext context) {
+
+		String text = context.getSearchFor();
+		if (text==null || text.length()==0) {
+			return false;
+		}
 
 		// Be smart about what position we're "starting" at.  We don't want
 		// to find a match in the currently selected text (if any), so we
 		// start searching AFTER the selection if searching forward, and
 		// BEFORE the selection if searching backward.
 		Caret c = textArea.getCaret();
+		boolean forward = context.getSearchForward();
 		int start = forward ? Math.max(c.getDot(), c.getMark()) :
 						Math.min(c.getDot(), c.getMark());
 
@@ -90,9 +82,9 @@ public class SearchEngine {
 		if (findIn==null || findIn.length()==0) return false;
 
 		// Find the next location of the text we're searching for.
-		if (regex==false) {
+		if (!context.isRegularExpression()) {
 			int pos = getNextMatchPos(text, findIn, forward,
-											matchCase, wholeWord);
+								context.getMatchCase(), context.getWholeWord());
 			findIn = null; // May help garbage collecting.
 			if (pos!=-1) {
 				// Without this, if JTextArea isn't in focus, selection
@@ -107,8 +99,8 @@ public class SearchEngine {
 			// Regex matches can have varying widths.  The returned point's
 			// x- and y-values represent the start and end indices of the
 			// match in findIn.
-			Point regExPos = getNextMatchPosRegEx(text, findIn,
-									forward, matchCase, wholeWord);
+			Point regExPos = getNextMatchPosRegEx(text, findIn, forward,
+								context.getMatchCase(), context.getWholeWord());
 			findIn = null; // May help garbage collecting.
 			if (regExPos!=null) {
 				// Without this, if JTextArea isn't in focus, selection
@@ -128,33 +120,57 @@ public class SearchEngine {
 	}
 
 
+	/**
+	 * Returns a <code>CharSequence</code> for a text area that doesn't make a
+	 * copy of its contents for iteration.  This conserves memory but is likely
+	 * just a tad slower.
+	 *
+	 * @param textArea The text area whose document is the basis for the
+	 *        <code>CharSequence</code>.
+	 * @param start The starting offset of the sequence (or ending offset if
+	 *        <code>forward</code> is <code>false</code>).
+	 * @param forward Whether we're searching forward or backward.
+	 * @return The character sequence.
+	 */
+	private static CharSequence getFindInCharSequence(RTextArea textArea,
+			int start, boolean forward) {
+		RDocument doc = (RDocument)textArea.getDocument();
+		int csStart = 0;
+		int csEnd = 0;
+		if (forward) {
+			csStart = start;
+			csEnd = doc.getLength();
+		}
+		else {
+			csStart = 0;
+			csEnd = start;
+		}
+		return new RDocumentCharSequence(doc, csStart, csEnd);
+	}
+
+
 	/**
 	 * Returns the text in which to search, as a string.  This is used
 	 * internally to grab the smallest buffer possible in which to search.
 	 */
-	protected static String getFindInText(JTextArea textArea, int start,
+	private static String getFindInText(JTextArea textArea, int start,
 									boolean forward) {
 
 		// Be smart about the text we grab to search in.  We grab more than
-		// a single line because our searches can return multiline results.
+		// a single line because our searches can return multi-line results.
 		// We copy only the chars that will be searched through.
 		String findIn = null;
-		if (forward) {
-			try {
+		try {
+			if (forward) {
 				findIn = textArea.getText(start,
 							textArea.getDocument().getLength()-start);
-			} catch (BadLocationException ble) {
-				// Never happens; findIn will be null anyway.
-				ble.printStackTrace();
 			}
-		}
-		else { // backward
-			try {
+			else { // backward
 				findIn = textArea.getText(0, start);
-			} catch (BadLocationException ble) {
-				// Never happens; findIn will be null anyway.
-				ble.printStackTrace();
 			}
+		} catch (BadLocationException ble) {
+			// Never happens; findIn will be null anyway.
+			ble.printStackTrace();
 		}
 
 		return findIn;
@@ -190,7 +206,8 @@ public class SearchEngine {
 	 *         an invalid group (less than zero or greater than the number of
 	 *         groups matched).
 	 */
-	protected static List getMatches(Matcher m, String replaceStr) {
+	@SuppressWarnings({ "rawtypes", "unchecked" })
+	private static List getMatches(Matcher m, String replaceStr) {
 		ArrayList matches = new ArrayList();
 		while (m.find()) {
 			Point loc = new Point(m.start(), m.end());
@@ -209,7 +226,9 @@ public class SearchEngine {
 	/**
 	 * Searches <code>searchIn</code> for an occurrence of
 	 * <code>searchFor</code> either forwards or backwards, matching
-	 * case or not.
+	 * case or not.<p>
+	 *
+	 * Most clients will have no need to call this method directly.
 	 *
 	 * @param searchFor The string to look for.
 	 * @param searchIn The string to search in.
@@ -222,8 +241,6 @@ public class SearchEngine {
 	 *        don't count as matches.
 	 * @return The starting position of a match, or <code>-1</code> if no
 	 *         match was found.
-	 * @see #getNextMatchPosImpl
-	 * @see #getNextMatchPosRegEx
 	 */
 	public static final int getNextMatchPos(String searchFor, String searchIn,
 								boolean forward, boolean matchCase,
@@ -259,7 +276,7 @@ public class SearchEngine {
 	 * @return The location of the next match, or <code>-1</code> if no
 	 *         match was found.
 	 */
-	protected static final int getNextMatchPosImpl(String searchFor,
+	private static final int getNextMatchPosImpl(String searchFor,
 								String searchIn, boolean goForward,
 								boolean matchCase, boolean wholeWord) {
 
@@ -312,7 +329,7 @@ public class SearchEngine {
 	 *         regular expression.
 	 * @see #getNextMatchPos
 	 */
-	public static Point getNextMatchPosRegEx(String regEx,
+	private static Point getNextMatchPosRegEx(String regEx,
 							CharSequence searchIn, boolean goForward,
 							boolean matchCase, boolean wholeWord) {
 		return (Point)getNextMatchPosRegExImpl(regEx, searchIn, goForward,
@@ -350,11 +367,15 @@ public class SearchEngine {
 	 *         groups matched).
 	 * @see #getNextMatchPos
 	 */
-	protected static Object getNextMatchPosRegExImpl(String regEx,
+	private static Object getNextMatchPosRegExImpl(String regEx,
 							CharSequence searchIn, boolean goForward,
 							boolean matchCase, boolean wholeWord,
 							String replaceStr) {
 
+		if (wholeWord) {
+			regEx = "\\b" + regEx + "\\b";
+		}
+
 		// Make a pattern that takes into account whether or not to match case.
 		int flags = Pattern.MULTILINE; // '^' and '$' are done per line.
 		flags |= matchCase ? 0 : (Pattern.CASE_INSENSITIVE|Pattern.UNICODE_CASE);
@@ -363,101 +384,28 @@ public class SearchEngine {
 		// Make a Matcher to find the regEx instances.
 		Matcher m = pattern.matcher(searchIn);
 
-		/*
-		 * Our algorithm is broken into four cases:
-		 *   1. Forward search, not whole-word: Just take first match found.
-		 *   2. Forward search, whole-word: Loop until the first whole-word
-		 *      match is found.
-		 *   3. Backward search, not whole-word.  Find all matches first
-		 *      (must do this since we can't search for regexes backwards),
-		 *      and return last match found.
-		 *   4. Backward search, whole-word.  Find all matches first, then
-		 *      loop through them backwards until the first (i.e., the last!)
-		 *      whole-word match is found.
-		 */
-
-		// If this is a forward-direction search...
+		// Search forwards
 		if (goForward) {
-
-			// 1. Forward search, not whole word => easy.  Just return
-			// the first match found.
-			if (!wholeWord) {
-				if (m.find()) {
-					if (replaceStr==null) { // Find, not replace.
-						return new Point(m.start(), m.end());
-					}
-					else { // Replace.
-						return new RegExReplaceInfo(m.group(0),
-								m.start(), m.end(),
-								getReplacementText(m, replaceStr));
-					}
-				}
-			}
-
-			// 2. Forward search, whole word => just okay.  Find and look at
-			// matches one at a time until you find one that's "whole word."
-			else {
-				while (m.find()) {
-					Point loc = new Point(m.start(), m.end());
-					if (isWholeWord(searchIn, loc.x,loc.y-loc.x)) {
-						if (replaceStr==null) { // Find, not replace.
-							return loc;
-						}
-						else { // Replace.
-							return new RegExReplaceInfo(m.group(0),
-								loc.x, loc.y,
-								getReplacementText(m, replaceStr));
-						}
-					}
-				}
-			}
-
-		} // End of if (goForward).
-
-		// If this is a backward-direction search...
-		else {
-
-			// Get some variables ready.
-			List matches = getMatches(m, replaceStr);
-			if (matches.isEmpty())
-				return null;
-			int pos = matches.size() - 1;
-
-			// 3. If they're not looking for a "whole word" just return
-			// the first (i.e., last) match.
-			if (wholeWord==false) {
+			if (m.find()) {
 				if (replaceStr==null) { // Find, not replace.
-					return /*(Point)*/matches.get(pos);
-				}
-				else { // Replace.
-					return /*(RegExReplaceInfo)*/matches.get(pos);
+					return new Point(m.start(), m.end());
 				}
+				// Otherwise, replace
+				return new RegExReplaceInfo(m.group(0),
+						m.start(), m.end(),
+						getReplacementText(m, replaceStr));
 			}
-
-			// 4. Otherwise, go through the matches last-to-first.
-			while (pos>=0) {
-				Object matchObj = matches.get(pos);
-				if (replaceStr==null) { // Find, not replace.
-					Point loc = (Point)matchObj;
-					if (isWholeWord(searchIn, loc.x,loc.y-loc.x)) {
-						return matchObj;
-					}
-				}
-				else { // Replace.
-					RegExReplaceInfo info = (RegExReplaceInfo)matchObj;
-					int x = info.getStartIndex();
-					int y = info.getEndIndex();
-					if (isWholeWord(searchIn, x,y-x)) {
-						return matchObj;
-					}
-				}
-				pos--;
-			}
-
 		}
 
-		// If we didn't find a match after all that, return null.
-		return null;
+		// Search backwards
+		else {
+			List<?> matches = getMatches(m, replaceStr);
+			if (!matches.isEmpty()) {
+				return matches.get(matches.size()-1);
+			}
+		}
+
+		return null; // No match found
 
 	}
 
@@ -466,39 +414,31 @@ public class SearchEngine {
 	 * Returns information on how to implement a regular expression "replace"
 	 * action in the specified text with the specified replacement string.
 	 *
-	 * @param regEx The regular expression to look for.
 	 * @param searchIn The string to search in.
-	 * @param goForward Whether to search forward.  If <code>false</code>,
-	 *        search backward.
-	 * @param matchCase Whether or not to do a case-sensitive search for
-	 *        <code>regEx</code>.
-	 * @param wholeWord If <code>true</code>, <code>regEx</code> occurrences
-	 *        embedded in longer words in <code>searchIn</code> don't count as
-	 *        matches.
-	 * @param replacement A template for the replacement string (e.g., this
-	 *        can contain <code>\t</code> and <code>\n</code> to mean tabs
-	 *        and newlines, respectively, as well as group references
-	 *        <code>$n</code>).
+	 * @param context The search options.
 	 * @return A <code>RegExReplaceInfo</code> object describing how to
 	 *         implement the replace.
-	 * @throws PatternSyntaxException If <code>regEx</code> is an invalid
-	 *         regular expression.
-	 * @throws IndexOutOfBoundsException If <code>replacement</code> references
-	 *         an invalid group (less than zero or greater than the number of
+	 * @throws PatternSyntaxException If the search text is an invalid regular
+	 *         expression.
+	 * @throws IndexOutOfBoundsException If the replacement text references an
+	 *         invalid group (less than zero or greater than the number of
 	 *         groups matched).
 	 * @see #getNextMatchPos
 	 */
-	protected static RegExReplaceInfo getRegExReplaceInfo(String regEx,
-									String searchIn, boolean goForward,
-									boolean matchCase, boolean wholeWord,
-									String replacement) {
+	private static RegExReplaceInfo getRegExReplaceInfo(CharSequence searchIn,
+										SearchContext context) {
 		// Can't pass null to getNextMatchPosRegExImpl or it'll think
 		// you're doing a "find" operation instead of "replace, and return a
 		// Point.
+		String replacement = context.getReplaceWith();
 		if (replacement==null) {
 			replacement = "";
 		}
-		return (RegExReplaceInfo)getNextMatchPosRegExImpl(regEx, searchIn,
+		String regex = context.getSearchFor();
+		boolean goForward = context.getSearchForward();
+		boolean matchCase = context.getMatchCase();
+		boolean wholeWord = context.getWholeWord();
+		return (RegExReplaceInfo)getNextMatchPosRegExImpl(regex, searchIn,
 						goForward, matchCase, wholeWord, replacement);
 	}
 
@@ -506,7 +446,15 @@ public class SearchEngine {
 	/**
 	 * Called internally by <code>getMatches()</code>.  This method assumes
 	 * that the specified matcher has just found a match, and that you want
-	 * to get the string with which to replace that match.
+	 * to get the string with which to replace that match.<p>
+	 *
+	 * Escapes simply insert the escaped character, except for <code>\n</code>
+	 * and <code>\t</code>, which insert a newline and tab respectively.
+	 * Substrings of the form <code>$\d+</code> are considered to be matched
+	 * groups.  To include a literal dollar sign in your template, escape it
+	 * (i.e. <code>\$</code>).<p>
+	 *
+	 * Most clients will have no need to call this method directly.
 	 *
 	 * @param m The matcher.
 	 * @param template The template for the replacement string.  For example,
@@ -526,7 +474,7 @@ public class SearchEngine {
 
 		// Process substitution string to replace group references with groups
 		int cursor = 0;
-		StringBuffer result = new StringBuffer();
+		StringBuilder result = new StringBuilder();
 
 		while (cursor < template.length()) {
 	
@@ -600,9 +548,8 @@ public class SearchEngine {
 
 	/**
 	 * Returns whether the characters on either side of
-	 * <code>substr(searchIn,startPos,startPos+searchStringLength)</code>
-	 * are whitespace.  While this isn't the best definition of "whole word",
-	 * it's the one we're going to use for now.
+	 * <code>substr(searchIn, startPos, startPos+searchStringLength)</code>
+	 * are <em>not</em> letters or digits.
 	 */
 	private static final boolean isWholeWord(CharSequence searchIn,
 											int offset, int len) {
@@ -631,7 +578,7 @@ public class SearchEngine {
 	 *        document (<code>false</code> means backward).
 	 * @return The new dot and mark position.
 	 */
-	protected static int makeMarkAndDotEqual(JTextArea textArea,
+	private static int makeMarkAndDotEqual(JTextArea textArea,
 										boolean forward) {
 		Caret c = textArea.getCaret();
 		int val = forward ? Math.min(c.getDot(), c.getMark()) :
@@ -647,25 +594,18 @@ public class SearchEngine {
 	 * the specified replacement string.
 	 *
 	 * @param textArea The text area in which to search.
-	 * @param toFind The regular expression to search for.
-	 * @param replaceWith The string to replace the found regex with.
-	 * @param forward Whether to search forward from the caret position
-	 *        or backward from it.
-	 * @param matchCase Whether the search should be case-sensitive.
-	 * @param wholeWord Whether there should be spaces or tabs on either
-	 *        side of the match.
+	 * @param context What to search for and all search options.
 	 * @return Whether a match was found (and thus replaced).
-	 * @throws PatternSyntaxException If <code>toFind</code> is not a
-	 *         valid regular expression.
-	 * @throws IndexOutOfBoundsException If <code>replaceWith</code> references
-	 *         an invalid group (less than zero or greater than the number of
-	 *         groups matched).
-	 * @see #replace
-	 * @see #find
+	 * @throws PatternSyntaxException If this is a regular expression search
+	 *         but the search text is an invalid regular expression.
+	 * @throws IndexOutOfBoundsException If this is a regular expression search
+	 *         but the replacement text references an invalid group (less than
+	 *         zero or greater than the number of groups matched).
+	 * @see #replace(RTextArea, SearchContext)
+	 * @see #find(JTextArea, SearchContext)
 	 */
-	protected static boolean regexReplace(JTextArea textArea, String toFind,
-				String replaceWith, boolean forward, boolean matchCase,
-				boolean wholeWord) throws PatternSyntaxException {
+	private static boolean regexReplace(RTextArea textArea,
+			SearchContext context) throws PatternSyntaxException {
 
 		// Be smart about what position we're "starting" at.  For example,
 		// if they are searching backwards and there is a selection such that
@@ -674,15 +614,14 @@ public class SearchEngine {
 		// selection.  So, in that case we start at the beginning of the
 		// selection.
 		Caret c = textArea.getCaret();
+		boolean forward = context.getSearchForward();
 		int start = makeMarkAndDotEqual(textArea, forward);
 
-		String findIn = getFindInText(textArea, start, forward);
+		CharSequence findIn = getFindInCharSequence(textArea, start, forward);
 		if (findIn==null) return false;
 
 		// Find the next location of the text we're searching for.
-		RegExReplaceInfo info = getRegExReplaceInfo(toFind, findIn,
-											forward, matchCase,
-											wholeWord, replaceWith);
+		RegExReplaceInfo info = getRegExReplaceInfo(findIn, context);
 
 		findIn = null; // May help garbage collecting.
 
@@ -718,37 +657,30 @@ public class SearchEngine {
 	 * specified replacement string.
 	 *
 	 * @param textArea The text area in which to search.
-	 * @param toFind The text/regular expression  to search for.
-	 * @param replaceWith The string to replace the found text with.
-	 * @param forward Whether to search forward from the caret position or
-	 *        backward from it.
-	 * @param matchCase Whether the search should be case-sensitive.
-	 * @param wholeWord Whether there should be spaces or tabs on either
-	 *        side of the match.
-	 * @param regex Whether or not this is a regular expression search.
+	 * @param context What to search for and all search options.
 	 * @return Whether a match was found (and thus replaced).
-	 * @throws PatternSyntaxException If <code>regex</code> is
-	 *         <code>true</code> but <code>toFind</code> is not a valid
-	 *         regular expression.
-	 * @throws IndexOutOfBoundsException If <code>regex</code> is
-	 *         <code>true</code> and <code>replaceWith</code> references
-	 *         an invalid group (less than zero or greater than the number
-	 *         of groups matched).
-	 * @see #regexReplace
-	 * @see #find
+	 * @throws PatternSyntaxException If this is a regular expression search
+	 *         but the search text is an invalid regular expression.
+	 * @throws IndexOutOfBoundsException If this is a regular expression search
+	 *         but the replacement text references an invalid group (less than
+	 *         zero or greater than the number of groups matched).
+	 * @see #replaceAll(RTextArea, SearchContext)
+	 * @see #find(JTextArea, SearchContext)
 	 */
-	public static boolean replace(RTextArea textArea, String toFind,
-				String replaceWith, boolean forward, boolean matchCase,
-				boolean wholeWord, boolean regex)
+	public static boolean replace(RTextArea textArea, SearchContext context)
 									throws PatternSyntaxException {
 
+		String toFind = context.getSearchFor();
+		if (toFind==null || toFind.length()==0) {
+			return false;
+		}
+
 		textArea.beginAtomicEdit();
 		try {
 
 			// Regular expression replacements have their own method.
-			if (regex) {
-				return regexReplace(textArea, toFind, replaceWith, forward,
-								matchCase, wholeWord);
+			if (context.isRegularExpression()) {
+				return regexReplace(textArea, context);
 			}
 
 			// Plain text search.  If we find it, replace it!
@@ -757,9 +689,9 @@ public class SearchEngine {
 			// to replace, then click "Replace" to replace the current
 			// selection. Since our find() method searches from an endpoint of
 			// the selection, we must remove the selection to work properly.
-			makeMarkAndDotEqual(textArea, forward);
-			if (find(textArea, toFind, forward, matchCase, wholeWord, false)) {
-				textArea.replaceSelection(replaceWith);
+			makeMarkAndDotEqual(textArea, context.getSearchForward());
+			if (find(textArea, context)) {
+				textArea.replaceSelection(context.getReplaceWith());
 				return true;
 			}
 
@@ -777,67 +709,37 @@ public class SearchEngine {
 	 * the specified document with the specified replacement.
 	 *
 	 * @param textArea The text area in which to search.
-	 * @param toFind The text/regular expression  to search for.
-	 * @param replaceWith The string to replace the found text with.
-	 * @param matchCase Whether the search should be case-sensitive.
-	 * @param wholeWord Whether there should be spaces or tabs on either
-	 *        side of the match.
-	 * @param regex Whether or not this is a regular expression search.
+	 * @param context What to search for and all search options.
 	 * @return The number of replacements done.
-	 * @throws PatternSyntaxException If <code>regex</code> is
-	 *         <code>true</code> and <code>toFind</code> is an invalid
-	 *         regular expression.
-	 * @throws IndexOutOfBoundsException If <code>replaceWith</code> references
-	 *         an invalid group (less than zero or greater than the number of
-	 *         groups matched).
-	 * @see #replace
-	 * @see #regexReplace
-	 * @see #find
+	 * @throws PatternSyntaxException If this is a regular expression search
+	 *         but the replacement text is an invalid regular expression.
+	 * @throws IndexOutOfBoundsException If this is a regular expression search
+	 *         but the replacement text references an invalid group (less than
+	 *         zero or greater than the number of groups matched).
+	 * @see #replace(RTextArea, SearchContext)
+	 * @see #find(JTextArea, SearchContext)
 	 */
-	public static int replaceAll(RTextArea textArea, String toFind,
-							String replaceWith, boolean matchCase,
-							boolean wholeWord, boolean regex)
+	public static int replaceAll(RTextArea textArea, SearchContext context)
 									throws PatternSyntaxException {
 
+		context.setSearchForward(true); // Replace all always searches forward
+		String toFind = context.getSearchFor();
+		if (toFind==null || toFind.length()==0) {
+			return 0;
+		}
+
 		int count = 0;
 
 		textArea.beginAtomicEdit();
 		try {
-
-			if (regex) {
-
-				if (replaceWith==null) {
-					replaceWith = ""; // Needed by getReplacementText() below.
-				}
-
-				int oldOffs = textArea.getCaretPosition();
-				textArea.setCaretPosition(0);
-				int flags = Pattern.MULTILINE; // '^' and '$' are done per line.
-				flags |= matchCase ? 0 :
-							Pattern.CASE_INSENSITIVE|Pattern.UNICODE_CASE;
-				Pattern p = Pattern.compile(toFind, flags);
-				while (SearchEngine.find(textArea, toFind, true, matchCase,
-									wholeWord, true)) {
-					Matcher m = p.matcher(textArea.getSelectedText());
-					String replacement = getReplacementText(m, replaceWith);
-					textArea.replaceSelection(replacement);
-					count++;
-				}
-				if (count==0) { // If nothing was found, don't move the caret.
-					textArea.setCaretPosition(oldOffs);
-				}
-
+			int oldOffs = textArea.getCaretPosition();
+			textArea.setCaretPosition(0);
+			while (SearchEngine.replace(textArea, context)) {
+				count++;
 			}
-
-			else { // Non-regular expression search.
-				textArea.setCaretPosition(0);
-				while (SearchEngine.find(textArea, toFind, true, matchCase,
-									wholeWord, false)) {
-					textArea.replaceSelection(replaceWith);
-					count++;
-				}
+			if (count==0) { // If nothing was found, don't move the caret.
+				textArea.setCaretPosition(oldOffs);
 			}
-
 		} finally {
 			textArea.endAtomicEdit();
 		}
@@ -859,6 +761,16 @@ public class SearchEngine {
 	private static void selectAndPossiblyCenter(JTextArea textArea, int start,
 												int end) {
 
+		boolean foldsExpanded = false;
+		if (textArea instanceof RSyntaxTextArea) {
+			RSyntaxTextArea rsta = (RSyntaxTextArea)textArea;
+			FoldManager fm = rsta.getFoldManager();
+			if (fm.isCodeFoldingSupportedAndEnabled()) {
+				foldsExpanded = fm.ensureOffsetNotInClosedFold(start);
+				foldsExpanded |= fm.ensureOffsetNotInClosedFold(end);
+			}
+		}
+
 		textArea.setSelectionStart(start);
 		textArea.setSelectionEnd(end);
 
@@ -882,7 +794,7 @@ public class SearchEngine {
 
 		// If the new selection is already in the view, don't scroll,
 		// as that is visually jarring.
-		if (visible.contains(r)) {
+		if (!foldsExpanded && visible.contains(r)) {
 			textArea.setSelectionStart(start);
 			textArea.setSelectionEnd(end);
 			return;
